#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Callable, Union

import numpy as np
import pandas as pd
from typing_extensions import Literal

from saqc.constants import UNFLAGGED
from saqc.core import register
from saqc.core.history import History
from saqc.lib.docs import DOC_TEMPLATES
from saqc.lib.tools import filterKwargs, getFreqDelta, isflagged
from saqc.lib.ts_operators import aggregate2Freq

if TYPE_CHECKING:
    from saqc import SaQC


METHOD2ARGS = {
    "inverse_fshift": ("backward", pd.Timedelta),
    "inverse_bshift": ("forward", pd.Timedelta),
    "inverse_nshift": ("nearest", lambda x: pd.Timedelta(x) / 2),
    "inverse_fagg": ("bfill", pd.Timedelta),
    "inverse_bagg": ("ffill", pd.Timedelta),
    "inverse_nagg": ("nearest", lambda x: pd.Timedelta(x) / 2),
    "match": (None, lambda _: "0min"),
}


class ResamplingMixin:
    @register(mask=["field"], demask=[], squeeze=[])
    def linear(
        self: "SaQC",
        field: str,
        freq: str,
        **kwargs,
    ) -> "SaQC":
        """
        A method to "regularize" data by interpolating linearly the data at regular timestamp.

        A series of data is considered "regular", if it is sampled regularly (= having uniform sampling rate).

        Interpolated values will get assigned the worst flag within freq-range.

        Note, that the data only gets interpolated at those (regular) timestamps, that have a valid (existing and
        not-na) datapoint preceeding them and one succeeding them within freq range.
        Regular timestamp that do not suffice this condition get nan assigned AND The associated flag will be of value
        ``UNFLAGGED``.

        Parameters
        ----------
        freq :
            An offset string. The frequency of the grid you want to interpolate your data at.
        """
        reserved = ["method", "order", "limit", "downgrade"]
        kwargs = filterKwargs(kwargs, reserved)
        return self.interpolateIndex(field, freq, "time", **kwargs)

    @register(mask=["field"], demask=[], squeeze=[])
    def shift(
        self: "SaQC",
        field: str,
        freq: str,
        method: Literal["fshift", "bshift", "nshift"] = "nshift",
        **kwargs,
    ) -> "SaQC":
        """
        Shift data points and flags to a regular frequency grid.

        Parameters
        ----------
        freq :
            Offset string. Sampling rate of the target frequency.

        method :
            Method to propagate values:

            * 'nshift' : shift grid points to the nearest time stamp in the range = +/- 0.5 * ``freq``
            * 'bshift' : shift grid points to the first succeeding time stamp (if any)
            * 'fshift' : shift grid points to the last preceeding time stamp (if any)
        """
        warnings.warn(
            f"""
            The method `shift` is deprecated and will be removed with version 2.6 of saqc.
            To achieve the same behavior please use:
            `qc.align(field={field}, freq={freq}. method={method})`
            """,
            DeprecationWarning,
        )

        return self.align(field=field, freq=freq, method=method, **kwargs)

    @register(mask=["field"], demask=[], squeeze=[])
    def resample(
        self: "SaQC",
        field: str,
        freq: str,
        func: Callable[[pd.Series], pd.Series] = np.mean,
        method: Literal["fagg", "bagg", "nagg"] = "bagg",
        maxna: int | None = None,
        maxna_group: int | None = None,
        **kwargs,
    ) -> "SaQC":
        """
        Resample data points and flags to a regular frequency.

        The data will be sampled to regular (equidistant) timestamps.
        Sampling intervals therefore get aggregated with a function, specified by
        ``func``, the result is projected to the new timestamps using
        ``method``. The following methods are available:

        * ``'nagg'``: all values in the range (+/- `freq`/2) of a grid point get
            aggregated with func and assigned to it.
        * ``'bagg'``: all values in a sampling interval get aggregated with func and
            the result gets assigned to the last grid point.
        * ``'fagg'``: all values in a sampling interval get aggregated with func and
            the result gets assigned to the next grid point.

        Note
        ----
        For perfomance reasons, ``func`` will be mapped to pandas.resample methods,
        if possible. However, for this to work, functions need an initialized
        ``__name__`` attribute, holding the function's name. Furthermore, you should
        not pass numpys nan-functions (``nansum``, ``nanmean``,...) because they
        cannot be optimised and the handling of ``NaN`` is already taken care of.

        Parameters
        ----------
        freq :
            Offset string. Sampling rate of the target frequency grid.

        func : default mean
            Aggregation function. See notes for performance considerations.

        method :
            Specifies which intervals to be aggregated for a certain timestamp. (preceding,
            succeeding or "surrounding" interval). See description above for more details.

        maxna :
            Maximum number of allowed ``NaN``s in a resampling interval. If exceeded, the
            entire interval is filled with ``NaN``.

        maxna_group :
            Same as `maxna` but for consecutive NaNs.
        """

        datcol = self._data[field]

        if datcol.empty:
            # see for #GL-374
            datcol = pd.Series(index=pd.DatetimeIndex([]), dtype=datcol.dtype)

        datcol = aggregate2Freq(
            datcol,
            method,
            freq,
            func,
            fill_value=np.nan,
            max_invalid_total=maxna,
            max_invalid_consec=maxna_group,
        )

        kws = dict(
            method=method,
            freq=freq,
            agg_func=max,
            fill_value=np.nan,
            max_invalid_total=maxna,
            max_invalid_consec=maxna_group,
        )

        history = self._flags.history[field].apply(
            index=datcol.index,
            func=aggregate2Freq,
            func_kws=kws,
        )
        meta = {
            "func": "resample",
            "args": (),
            "kwargs": {
                "freq": freq,
                "func": func,
                "method": method,
                "maxna": maxna,
                "maxna_group": maxna_group,
                **kwargs,
            },
        }
        flagcol = pd.Series(UNFLAGGED, index=history.index)
        history.append(flagcol, meta)

        self._data[field] = datcol
        self._flags.history[field] = history
        return self

    @register(
        mask=[],
        demask=[],
        squeeze=[],
        handles_target=True,  # target is mandatory in func, so its allowed
        docstring={"target": DOC_TEMPLATES["target"]},
    )
    def concatFlags(
        self: "SaQC",
        field: str,
        target: str | None = None,
        method: Literal[
            "inverse_fagg",
            "inverse_bagg",
            "inverse_nagg",
            "inverse_fshift",
            "inverse_bshift",
            "inverse_nshift",
            "inverse_interpolation",
            "match",
            "auto",
        ] = "match",
        freq: str | None = None,
        drop: bool = False,
        squeeze: bool = False,
        overwrite: bool = False,
        **kwargs,
    ) -> "SaQC":
        """
        Project flags/history of :py:attr:`field` to :py:attr:`target` and adjust to the frequeny grid
        of :py:attr:`target` by 'undoing' former interpolation, shifting or resampling operations

        Note
        ----
        To undo or backtrack resampling, shifting or interpolation operations, use the
        associated inversion method (e.g. to undo a former interpolation use
        ``method="inverse_interpolation"``).

        Parameters
        ----------
        method :
            Method to project the flags of :py:attr:`field` the flags to :py:attr:`target`:

           * ``'auto'``: inverse the last alignment/resampling operations
           * ``'inverse_nagg'``: project a flag of :py:attr:`field` to all timestamps of
             :py:attr:`target` within the range +/- :py:attr:`freq`/2.
           * ``'inverse_bagg'``: project a flag of :py:attr:`field` to all preceeding timestamps
             of :py:attr:`target` within the range :py:attr:`freq`
           * ``'inverse_fagg'``: project a flag of :py:attr:`field` to all succeeding timestamps
             of :py:attr:`target` within the range :py:attr:`freq`
           * ``'inverse_interpolation'`` - project a flag of :py:attr:`field` to all timestamps
             of :py:attr:`target` within the range +/- :py:attr:`freq`
           * ``'inverse_nshift'`` - project a flag of :py:attr:`field` to the neaerest timestamps
             in :py:attr:`target` within the range +/- :py:attr:`freq`/2
           * ``'inverse_bshift'`` - project a flag of :py:attr:`field` to nearest preceeding
             timestamps in :py:attr:`target`
           * ``'inverse_nshift'`` - project a flag of :py:attr:`field` to nearest succeeding
             timestamps in :py:attr:`target`
           * ``'match'`` - project a flag of :py:attr:`field` to all identical timestamps
             :py:attr:`target`

        freq :
            Projection range. If ``None`` the sampling frequency of :py:attr:`field` is used.

        drop :
            Remove :py:attr:`field` if ``True``

        squeeze :
            Squueze the history into a single column if ``True``. Function specific flag information is lost.

        overwrite :
            Overwrite existing flags if ``True``
        """

        if target is None:
            target = field

        flagscol = self._flags[field]
        target_datcol = self._data[target]
        target_flagscol = self._flags[target]

        if target_datcol.empty or flagscol.empty:
            return self

        dummy = pd.Series(np.nan, target_flagscol.index, dtype=float)

        if freq is None:
            freq = getFreqDelta(flagscol.index)
            if freq is None and not method == "match":
                raise ValueError(
                    'To project irregularly sampled data, either use method="match", or '
                    "pass custom projection range to freq parameter."
                )

        if method == "auto":
            stack = []
            for meta in self._flags.history[field].meta:
                func = meta["func"]
                meth = meta["kwargs"].get("method")
                if func in ("align", "resample"):
                    if meth[1:] in ("agg", "shift"):
                        stack.append(f"inverse_{meth}")
                    else:
                        stack.append("inverse_interpolation")
                elif func == "concatFlags":
                    stack.pop()
            if not stack:
                raise ValueError(
                    "unable to derive an inversion method, please specify an appropiate 'method'"
                )
            method = stack[-1]

        if method.endswith("interpolation"):
            ignore = _getChunkBounds(target_datcol, flagscol, freq)
            func = _inverseInterpolation
            func_kws = dict(freq=freq, chunk_bounds=ignore, target=dummy)

        elif method.endswith("agg"):
            projection_method = METHOD2ARGS[method][0]
            tolerance = METHOD2ARGS[method][1](freq)
            func = _inverseAggregation
            func_kws = dict(freq=tolerance, method=projection_method, target=dummy)

        elif method.endswith("shift"):
            drop_mask = target_datcol.isna() | isflagged(
                target_flagscol, kwargs["dfilter"]
            )
            projection_method = METHOD2ARGS[method][0]
            tolerance = METHOD2ARGS[method][1](freq)
            func = _inverseShift
            kws = dict(
                freq=tolerance,
                method=projection_method,
                drop_mask=drop_mask,
                target=dummy,
            )
            func_kws = {**kws, "fill_value": np.nan}

        elif method == "match":
            func = lambda x: x
            func_kws = {}

        else:
            raise ValueError(f"unknown method {method}")

        history = self._flags.history[field].apply(dummy.index, func, func_kws)
        if overwrite is False:
            mask = isflagged(self._flags[target], thresh=kwargs["dfilter"])
            history._hist[mask] = np.nan

        # append a dummy column
        meta = {
            "func": f"concatFlags",
            "args": (),
            "kwargs": {
                "field": field,
                "target": target,
                "method": method,
                "freq": freq,
                "drop": drop,
                "squeeze": squeeze,
                "overwrite": overwrite,
                **kwargs,
            },
        }

        if squeeze:
            flags = history.squeeze(raw=True)
            history = History(index=history.index)
        else:
            flags = pd.Series(np.nan, index=history.index, dtype=float)

        history.append(flags, meta)
        self._flags.history[target].append(history)

        if drop:
            return self.dropField(field=field)

        return self


def _getChunkBounds(target: pd.Series, flagscol: pd.Series, freq: str):
    chunk_end = target.reindex(flagscol.index, method="bfill", tolerance=freq)
    chunk_start = target.reindex(flagscol.index, method="ffill", tolerance=freq)
    ignore_flags = chunk_end.isna() | chunk_start.isna()
    return ignore_flags


def _inverseInterpolation(
    source: pd.Series, target: pd.Series, freq: str, chunk_bounds
) -> pd.Series:
    source = source.copy()
    if len(chunk_bounds) > 0:
        source[chunk_bounds] = np.nan
    backprojected = source.reindex(target.index, method="bfill", tolerance=freq)
    fwrdprojected = source.reindex(target.index, method="ffill", tolerance=freq)
    return pd.concat([backprojected, fwrdprojected], axis=1).max(axis=1)


def _inverseAggregation(
    source: Union[pd.Series, pd.DataFrame],
    target: Union[pd.Series, pd.DataFrame],
    freq: str,
    method: str,
):
    return source.reindex(target.index, method=method, tolerance=freq)


def _inverseShift(
    source: pd.Series,
    target: pd.Series,
    drop_mask: pd.Series,
    freq: str,
    method: str,
    fill_value,
) -> pd.Series:
    dtype = source.dtype

    target_drops = target[drop_mask]
    target = target[~drop_mask]
    flags_merged = pd.merge_asof(
        source,
        target.index.to_series(name="pre_index"),
        left_index=True,
        right_index=True,
        tolerance=freq,
        direction=method,
    )
    flags_merged.dropna(subset=["pre_index"], inplace=True)
    flags_merged = flags_merged.set_index(["pre_index"]).squeeze()
    target[flags_merged.index] = flags_merged.values

    # reinsert drops
    source = target.reindex(target.index.union(target_drops.index))
    source.loc[target_drops.index] = target_drops.values

    return source.fillna(fill_value).astype(dtype, copy=False)



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-

import pandas as pd
import pytest

from saqc import BAD, UNFLAGGED, SaQC
from saqc.core import DictOfSeries, initFlagsLike
from tests.common import initData


@pytest.fixture
def data():
    return initData(cols=1, start_date="2016-01-01", end_date="2018-12-31", freq="1D")


@pytest.fixture
def field(data):
    return data.columns[0]


@pytest.mark.parametrize("plot", [True, False])
@pytest.mark.parametrize("normalize", [True, False])
def test_flagPattern_dtw(plot, normalize):
    data = pd.Series(0, index=pd.date_range(start="2000", end="2001", freq="1d"))
    data.iloc[10:18] = [0, 5, 6, 7, 6, 8, 5, 0]
    pattern = data.iloc[10:18]

    data = DictOfSeries(data=data, pattern_data=pattern)
    flags = initFlagsLike(data, name="data")
    qc = SaQC(data, flags).flagPatternByDTW(
        "data",
        reference="pattern_data",
        plot=plot,
        normalize=normalize,
        flag=BAD,
    )

    assert all(qc.flags["data"].iloc[10:18] == BAD)
    assert all(qc.flags["data"].iloc[:9] == UNFLAGGED)
    assert all(qc.flags["data"].iloc[18:] == UNFLAGGED)



#!/usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Dict, Union

import numpy as np
import pandas as pd
import pytest

import tests.core.test_history as test_hist
from saqc import UNFLAGGED
from saqc.core import DictOfSeries, Flags, History

_arrays = [
    np.array([[]]),
    np.zeros((1, 1)),
    np.zeros((3, 4)),
    np.ones((3, 4)),
    np.ones((3, 4)) * np.nan,
    np.array(
        [
            [0, 0, 0, 0],
            [0, 1, 2, 3],
            [0, 1, 2, 3],
        ]
    ),
    np.array(
        [
            [0, 0, 0, 0],
            [0, 1, np.nan, 3],
            [0, 1, 2, 3],
        ]
    ),
]

testdata = []
for d in _arrays:
    columns = list("abcdefgh")[: d.shape[1]]
    df = pd.DataFrame(d, dtype=float, columns=columns)
    dis = DictOfSeries(df)
    di = {}
    di.update(df.items())
    testdata.append(df)
    testdata.append(di)
    testdata.append(dis)


def is_equal(f1, f2):
    """assert Flags instance equals other"""
    assert f1.columns.equals(f2.columns)
    for c in f1.columns:
        assert test_hist.is_equal(f1.history[c], f2.history[c])


@pytest.mark.parametrize("data", testdata)
@pytest.mark.parametrize("copy", [True, False])
def test_init(data, copy):
    flags = Flags(data, copy=copy)
    assert isinstance(flags, Flags)
    assert len(data.keys()) == len(flags)


@pytest.mark.parametrize("data", testdata)
@pytest.mark.parametrize("copy", [True, False])
def test_init_from_other(data, copy):
    first = Flags(data)
    second = Flags(first, copy=copy)
    assert isinstance(second, Flags)
    assert first is not second
    if copy:
        assert first._data is not second._data
    else:
        assert first._data is second._data


@pytest.mark.parametrize(
    "data,msg",
    [
        ({1: pd.Series([1.0])}, "column names must be string"),
        ({"a": pd.Series([1, 2], dtype=int)}, "dtype must be float"),
    ],
)
def test_init_raise_ValueError(data, msg):
    with pytest.raises(ValueError) as e:
        Flags(data)
    assert msg in str(e.value)


@pytest.mark.parametrize(
    "data,msg",
    [
        ({"a": [1, 2]}, "cannot init from 'dict' of 'list'"),
    ],
)
def test_init_raise_TypeError(data, msg):
    with pytest.raises(TypeError) as e:
        Flags(data)
    assert msg in str(e.value)


@pytest.mark.parametrize("data", testdata)
def test_copy(data: Union[pd.DataFrame, DictOfSeries, Dict[str, pd.Series]]):
    flags = Flags(data)
    shallow = flags.copy(deep=False)
    deep = flags.copy(deep=True)

    # checks

    for copy in [deep, shallow]:
        assert isinstance(copy, Flags)
        assert copy is not flags
        assert copy._data is not flags._data
        for c in copy.columns:
            assert copy._data[c] is not flags._data[c]
        is_equal(copy, flags)

    assert deep is not shallow
    is_equal(deep, shallow)

    # the underling series data is the same
    for c in shallow.columns:
        assert shallow._data[c].index is flags._data[c].index

    # the underling series data was copied
    for c in deep.columns:
        assert deep._data[c].index is not flags._data[c].index


@pytest.mark.parametrize("data", testdata)
def test_flags_history(data: Union[pd.DataFrame, DictOfSeries, Dict[str, pd.Series]]):
    flags = Flags(data)

    # get
    for c in flags.columns:
        hist = flags.history[c]
        assert isinstance(hist, History)
        assert len(hist) > 0

    # set
    for c in flags.columns:
        hist = flags.history[c]
        hlen = len(hist)
        hist.append(pd.Series(888.0, index=hist.index, dtype=float))
        flags.history[c] = hist
        assert isinstance(hist, History)
        assert len(hist) == hlen + 1


@pytest.mark.parametrize("data", testdata)
def test_get_flags(data: Union[pd.DataFrame, DictOfSeries, Dict[str, pd.Series]]):
    flags = Flags(data)

    for c in flags.columns:
        # check obvious
        var = flags[c]
        assert isinstance(var, pd.Series)
        assert not var.empty
        assert var.equals(flags._data[c].squeeze())

        # always a copy
        assert var is not flags[c]

        # in particular, a deep copy
        var[:] = 9999.0
        assert all(flags[c] != var)


@pytest.mark.parametrize("data", testdata)
def test_set_flags(data: Union[pd.DataFrame, DictOfSeries, Dict[str, pd.Series]]):
    flags = Flags(data)

    for c in flags.columns:
        var = flags[c]
        hlen = len(flags.history[c])
        new = pd.Series(9999.0, index=var.index, dtype=float)

        flags[c] = new
        assert len(flags.history[c]) == hlen + 1
        assert all(flags.history[c].squeeze() == 9999.0)
        assert all(flags.history[c].squeeze() == flags[c])

        # check if deep-copied correctly
        new[:] = 8888.0
        assert all(flags.history[c].squeeze() == 9999.0)

        # flags always overwrite former
        flags[c] = new
        assert len(flags.history[c]) == hlen + 2
        assert all(flags.history[c].squeeze() == 8888.0)
        assert all(flags.history[c].squeeze() == flags[c])

        # check if deep-copied correctly
        new[:] = 7777.0
        assert all(flags.history[c].squeeze() == 8888.0)


@pytest.mark.parametrize("data", testdata)
def test_set_flags_with_mask(
    data: Union[pd.DataFrame, DictOfSeries, Dict[str, pd.Series]]
):
    flags = Flags(data)

    for c in flags.columns:
        var = flags[c]
        mask = var == UNFLAGGED

        scalar = 222.0
        flags[mask, c] = scalar
        assert all(flags[c].loc[mask] == 222.0)
        assert all(flags[c].loc[~mask] != 222.0)

        # scalar without mask is not allowed, because
        # it holds to much potential to set the whole
        # column unintentionally.
        with pytest.raises(ValueError):
            flags[c] = 888.0

        vector = var.copy()
        vector[:] = 333.0
        flags[mask, c] = vector
        assert all(flags[c].loc[mask] == 333.0)
        assert all(flags[c].loc[~mask] != 333.0)

        # works with any that pandas eat, eg with numpy
        vector[:] = 444.0
        vector = vector.to_numpy()
        flags[mask, c] = vector
        assert all(flags[c].loc[mask] == 444.0)
        assert all(flags[c].loc[~mask] != 444.0)

        # test length miss-match (mask)
        if len(mask) > 1:
            wrong_len = mask[:-1]
            with pytest.raises(ValueError):
                flags[wrong_len, c] = vector

        # test length miss-match (value)
        if len(vector) > 1:
            wrong_len = vector[:-1]
            with pytest.raises(ValueError):
                flags[mask, c] = wrong_len


@pytest.mark.parametrize("data", testdata)
def test_set_flags_with_index(
    data: Union[pd.DataFrame, DictOfSeries, Dict[str, pd.Series]]
):
    flags = Flags(data)

    for c in flags.columns:
        var = flags[c]
        mask = var == UNFLAGGED
        index = mask[mask].index

        scalar = 222.0
        flags[index, c] = scalar
        assert all(flags[c].loc[mask] == 222.0)
        assert all(flags[c].loc[~mask] != 222.0)

        vector = var.copy()
        vector[:] = 333.0
        flags[index, c] = vector
        assert all(flags[c].loc[mask] == 333.0)
        assert all(flags[c].loc[~mask] != 333.0)

        # works with any that pandas eat, eg with numpy
        vector[:] = 444.0
        vector = vector.to_numpy()
        flags[index, c] = vector
        assert all(flags[c].loc[mask] == 444.0)
        assert all(flags[c].loc[~mask] != 444.0)

        # test length miss-match (value)
        if len(vector) > 1:
            wrong_len = vector[:-1]
            with pytest.raises(ValueError):
                flags[index, c] = wrong_len


def _validate_flags_equals_frame(flags, df):
    assert df.columns.equals(flags.columns)

    for c in flags.columns:
        assert df[c].index.equals(flags[c].index)
        assert df[c].equals(flags[c])  # respects nan's


@pytest.mark.parametrize("data", testdata)
def test_to_dios(data: Union[pd.DataFrame, DictOfSeries, Dict[str, pd.Series]]):
    flags = Flags(data)
    with pytest.deprecated_call():
        result = flags.toDios()
        assert isinstance(result, DictOfSeries)
        _validate_flags_equals_frame(flags, result)


@pytest.mark.parametrize("data", testdata)
def test_toFrame(data: Union[pd.DataFrame, DictOfSeries, Dict[str, pd.Series]]):
    flags = Flags(data)
    df = flags.toFrame()

    assert isinstance(df, pd.DataFrame)
    _validate_flags_equals_frame(flags, df)


@pytest.mark.parametrize(
    "columns",
    [["x", "y"], pd.Index(["x", "y"])],
)
def test_columns_setter(columns):
    flags = Flags(
        {"a": pd.Series([1, 2], dtype=float), "b": pd.Series([1, 2], dtype=float)}
    )
    flags.columns = columns
    for c in columns:
        assert c in flags.columns


@pytest.mark.parametrize(
    "columns,err",
    [
        ("foooo", TypeError),  # cannot cast to Index
        (pd.Index(["x", "x"]), TypeError),  # duplicates
        (pd.Index([1, 2]), TypeError),  # not string
        (pd.Index(["x", "y", "z"]), ValueError),  # wrong length
        (pd.Index(["x"]), ValueError),  # wrong length
    ],
)
def test_columns_setter_raises(columns, err):
    flags = Flags(
        {"a": pd.Series([1, 2], dtype=float), "b": pd.Series([1, 2], dtype=float)}
    )
    with pytest.raises(err):
        flags.columns = columns


@pytest.mark.parametrize(
    "data,key,expected",
    [
        (dict(a=[0, 1], b=[]), "a", pd.Series([0, 1], dtype=float)),
        (dict(a=[0, 1], b=[]), "b", pd.Series([], dtype=float)),
    ],
)
def test__getitem__scalar(data, key, expected):
    flags = Flags({k: pd.Series(v, dtype=float) for k, v in data.items()})
    result: pd.Series = flags[key]
    assert isinstance(result, pd.Series)
    assert result.equals(expected)
    # assert copying
    assert flags[key] is not flags[key]


@pytest.mark.parametrize(
    "data,key,expected",
    [
        (dict(a=[0, 1], b=[]), [], dict()),
        (dict(a=[0, 1], b=[]), ["a"], dict(a=[0, 1])),
        (dict(a=[0, 1], b=[]), ["a", "b"], dict(a=[0, 1], b=[])),
        (dict(a=[0, 1], b=[]), pd.Index([]), dict()),
        (dict(a=[0, 1], b=[]), pd.Index(["a"]), dict(a=[0, 1])),
        (dict(a=[0, 1], b=[]), pd.Index(["a", "b"]), dict(a=[0, 1], b=[])),
        (dict(a=[0, 1], b=[]), slice(None), dict(a=[0, 1], b=[])),
        (dict(a=[0, 1], b=[]), slice(0, 1), dict(a=[0, 1])),
        (dict(a=[0, 1], b=[]), slice(1, 99), dict(b=[])),
        (dict(a=[0, 1], b=[]), slice(5, 99), dict()),
    ],
)
def test__getitem__listlike_and_slice(data, key, expected):
    flags = Flags({k: pd.Series(v, dtype=float) for k, v in data.items()})
    result: Flags = flags[key]
    assert isinstance(result, Flags)
    # assert that a new Flags object was created
    assert flags[key] is not flags[key]
    # assert that internal data is copied
    if len(result):
        left = result._data[result.columns[0]]
        right = flags._data[result.columns[0]]
        assert left is not right

    expected = Flags({k: pd.Series(v, dtype=float) for k, v in expected.items()})
    is_equal(result, expected)



# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import saqc
from saqc.core import DictOfSeries


@pytest.mark.slow
def test_makeFig(tmp_path):
    # just testing for no errors to occure...
    data = DictOfSeries(
        data=pd.Series(
            np.linspace(0, 1000, 1000),
            pd.date_range("2000", "2001", periods=1000),
        )
    )
    d_saqc = saqc.SaQC(data)
    d_saqc = (
        d_saqc.flagRange("data", max=500)
        .flagRange("data", max=400)
        .flagRange("data", max=300)
    )

    # not interactive, no storing
    outfile = str(Path(tmp_path, "test.png"))  # the filesystem's temp dir

    d_saqc = d_saqc.plot(field="data", path=outfile)
    d_saqc = d_saqc.plot(field="data", path=outfile, history="valid", stats=True)
    d_saqc = d_saqc.plot(field="data", path=outfile, history="complete")
    d_saqc = d_saqc.plot(
        field="data", path=outfile, ax_kwargs={"ylabel": "data is data"}, stats=True
    )



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-

from __future__ import annotations

import operator
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from saqc import BAD
from saqc.core import flagging
from saqc.lib.rolling import removeRollingRamps
from saqc.lib.tools import getFreqDelta, statPass
from saqc.lib.ts_operators import varQC

if TYPE_CHECKING:
    from saqc import SaQC


class ConstantsMixin:
    @flagging()
    def flagConstants(
        self: "SaQC",
        field: str,
        thresh: float,
        window: int | str,
        min_periods: int = 2,
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Flag constant data values.

        Flags plateaus of constant data if their maximum total change in
        a rolling window does not exceed a certain threshold.

        Any interval of values y(t),...,y(t+n) is flagged, if:
         - (1): n > ``window``
         - (2): abs(y(t + i) - (t + j)) < `thresh`, for all i,j in [0, 1, ..., n]

        Parameters
        ----------
        thresh :
            Maximum total change allowed per window.

        window :
            Size of the moving window. This is the number of observations used
            for calculating the statistic. Each window will be a fixed size.
            If its an offset then this will be the time period of each window.
            Each window will be a variable sized based on the observations included
            in the time-period.
        """
        if not isinstance(window, (str, int)):
            raise TypeError("window must be offset string or int.")

        d: pd.Series = self._data[field]

        if not isinstance(window, int) and not pd.api.types.is_datetime64_any_dtype(
            d.index
        ):
            raise ValueError(
                f"A time based value for 'window' is only possible for variables "
                f"with a datetime based index, but variable '{field}' has an index "
                f"of dtype {d.index.dtype}. Use an integer window instead."
            )

        # min_periods=2 ensures that at least two non-nan values are present
        # in each window and also min() == max() == d[i] is not possible.
        min_periods = max(min_periods, 2)

        # 1. find starting points of consecutive constant values as a boolean mask
        # 2. fill the whole window with True's
        rolling = d.rolling(window=window, min_periods=min_periods)
        starting_points_mask = rolling.max() - rolling.min() <= thresh

        removeRollingRamps(starting_points_mask, window=window, inplace=True)

        # mimic forward rolling by roll over inverse [::-1]
        rolling = starting_points_mask[::-1].rolling(
            window=window, min_periods=min_periods
        )
        # mimic any()
        mask = (rolling.sum()[::-1] > 0) & d.notna()

        self._flags[mask, field] = flag
        return self

    @flagging()
    def flagByVariance(
        self: "SaQC",
        field: str,
        window: str,
        thresh: float,
        maxna: int | None = None,
        maxna_group: int | None = None,
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Flag low-variance data.

        Flags plateaus of constant data if the variance in a rolling window does not
        exceed a certain threshold.

        Any interval of values y(t),..y(t+n) is flagged, if:

        (1) n > `window`
        (2) variance(y(t),...,y(t+n) < `thresh`

        Parameters
        ----------
        window :
            Size of the moving window. This is the number of observations used
            for calculating the statistic. Each window will be a fixed size.
            If its an offset then this will be the time period of each window.
            Each window will be sized, based on the number of observations included
            in the time-period.

        thresh :
            Maximum total variance allowed per window.

        maxna :
            Maximum number of NaNs allowed in window.
            If more NaNs are present, the window is not flagged.

        maxna_group :
            Same as `maxna` but for consecutive NaNs.
        """
        dataseries = self._data[field]
        delta = getFreqDelta(dataseries.index)
        if not delta:
            raise IndexError("Timeseries irregularly sampled!")

        if maxna is None:
            maxna = np.inf

        if maxna_group is None:
            maxna_group = np.inf

        min_periods = int(np.ceil(pd.Timedelta(window) / pd.Timedelta(delta)))
        window = pd.Timedelta(window)
        to_set = statPass(
            dataseries,
            lambda x: varQC(x, maxna, maxna_group),
            window,
            thresh,
            operator.lt,
            min_periods=min_periods,
        )

        self._flags[to_set, field] = flag
        return self



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-
from __future__ import annotations

import collections
import functools
import itertools
import re
import warnings
from typing import Any, Callable, Collection, List, Sequence, TypeVar, Union

import numpy as np
import pandas as pd
from scipy import fft
from scipy.cluster.hierarchy import fcluster, linkage

from saqc.lib.types import CompT

T = TypeVar("T", str, float, int)


def assertScalar(name, value, optional=False):
    if optional and value is None:
        return
    if np.isscalar(value):
        return

    msg = f"'{name}' needs to be a scalar"
    if optional:
        msg += " or 'None'"
    raise ValueError(msg)


def toSequence(value: T | Sequence[T]) -> List[T]:
    if value is None:  # special case
        return [None]
    if isinstance(value, T.__constraints__):
        return [value]
    return list(value)


def squeezeSequence(value: Sequence[T]) -> Union[T, Sequence[T]]:
    if len(value) == 1:
        return value[0]
    return value


def periodicMask(dtindex, season_start, season_end, include_bounds):
    """
    This function generates date-periodic/seasonal masks from an index passed.

    For example you could mask all the values of an index, that are sampled in winter, or between 6 and 9 o'clock.
    See the examples section for more details.

    Parameters
    ----------
    dtindex : pandas.DatetimeIndex
        The index according to wich you want to generate a mask.
        (=resulting mask will be indexed with 'dtindex')
    season_start : str
        String denoting starting point of every period. Formally, it has to be a truncated instance of "mm-ddTHH:MM:SS".
        Has to be of same length as `season_end` parameter.
        See examples section below for some examples.
    season_end : str
        String denoting starting point of every period. Formally, it has to be a truncated instance of "mm-ddTHH:MM:SS".
        Has to be of same length as `season_end` parameter.
        See examples section below for some examples.
    include_bounds : boolean
        Wheather or not to include the mask defining bounds to the mask.

    Returns
    -------
    dfilter : pandas.Series[bool]
        A series, indexed with the input index and having value `False` for all the values that are to be masked.

    Examples
    --------
    The `season_start` and `season_end` parameters provide a conveniant way to generate seasonal / date-periodic masks.
    They have to be strings of the forms: "mm-ddTHH:MM:SS", "ddTHH:MM:SS" , "HH:MM:SS", "MM:SS" or "SS"
    (mm=month, dd=day, HH=hour, MM=minute, SS=second)
    Single digit specifications have to be given with leading zeros.
    `season_start` and `seas   on_end` strings have to be of same length (refer to the same periodicity)
    The highest date unit gives the period.
    For example:

    >>> season_start = "01T15:00:00"
    >>> season_end = "13T17:30:00"

    Will result in all values sampled between 15:00 at the first and  17:30 at the 13th of every month get masked

    >>> season_start = "01:00"
    >>> season_end = "04:00"

    All the values between the first and 4th minute of every hour get masked.

    >>> season_start = "01-01T00:00:00"
    >>> season_end = "01-03T00:00:00"

    Mask january and february of evcomprosed in theery year. masking is inclusive always, so in this case the mask will
    include 00:00:00 at the first of march. To exclude this one, pass:

    >>> season_start = "01-01T00:00:00"
    >>> season_end = "02-28T23:59:59"

    To mask intervals that lap over a seasons frame, like nights, or winter, exchange sequence of season start and
    season end. For example, to mask night hours between 22:00:00 in the evening and 06:00:00 in the morning, pass:

    >>> season_start = "22:00:00"
    >>> season_end = "06:00:00"

    When inclusive_selection="season", all above examples work the same way, only that you now
    determine wich values NOT TO mask (=wich values are to constitute the "seasons").
    """

    def _replaceBuilder(stamp):
        keys = ("second", "minute", "hour", "day", "month", "year")
        stamp_list = map(int, re.split(r"[-T:]", stamp)[::-1])
        stamp_kwargs = dict(zip(keys, stamp_list))

        def _replace(index):
            if len(index) == 0:
                return None
            if "day" in stamp_kwargs:
                stamp_kwargs["day"] = min(stamp_kwargs["day"], index[0].daysinmonth)
            out = index[0].replace(**stamp_kwargs)
            return out.strftime("%Y-%m-%dT%H:%M:%S")

        return _replace

    mask = pd.Series(True, index=dtindex)

    start_replacer = _replaceBuilder(season_start)
    end_replacer = _replaceBuilder(season_end)

    invert = False

    if pd.Timestamp(start_replacer(dtindex)) > pd.Timestamp(end_replacer(dtindex)):
        start_replacer, end_replacer = end_replacer, start_replacer
        include_bounds = not include_bounds
        invert = True

    if include_bounds:

        def _selector(x):
            x[start_replacer(x.index) : end_replacer(x.index)] = False
            return x

    else:

        def _selector(x):
            s = start_replacer(x.index)
            e = end_replacer(x.index)
            x[s:e] = False
            x[s:s] = True
            x[e:e] = True
            return x

    freq = "1" + "mmmhhhdddMMMYYY"[len(season_start)]
    out = mask.groupby(pd.Grouper(freq=freq)).transform(_selector)
    if invert:
        out = ~out
    return out


def isQuoted(string):
    return bool(re.search(r"'.*'|\".*\"", string))


def mutateIndex(index, old_name, new_name):
    pos = index.get_loc(old_name)
    index = index.drop(index[pos])
    index = index.insert(pos, new_name)
    return index


def estimateFrequency(
    index,
    delta_precision=-1,
    max_rate="10s",
    min_rate="1D",
    optimize=True,
    min_energy=0.2,
    max_freqs=10,
    bins=None,
):
    """
    Function to estimate the sampling rate of an index.

    The function comes with some optional overhead.
    The default options detect sampling rates from 10 seconds to 1 day with a 10 seconds precision
    for sampling rates below one minute, and a one minute precision for rates between 1 minute and
    one day.

    The function is designed to detect mixed sampling ratess as well as rate changes.
    In boh situations, all the sampling rates detected, are returned, together with their
    greatest common rate.

    Note, that there is a certain lower bound of index length,
    beneath which frequency leaking and Niquist limit take over and mess up the fourier
    transform.

    Parameters
    ----------
    index : {pandas.DateTimeIndex}
        The index of wich the sampling rates shall be estimated
    delta_precision : int, default -1
        determines detection precision. Precision equals: seconds*10**(-1-delta_precision).
        A too high precision attempt can lead to performance loss and doesnt necessarily result in
        a more precise result. Especially when the samples deviation from their mean rate
        is high compared to the delta_precision.
    max_rate : str, default "10s"
        Maximum rate that can be detected.
    min_rate : str, default "1D"
        Minimum detectable sampling rate.
    optimize : bool, default True
        Wheather or not to speed up fft application by zero padding the derived response series to
        an optimal length. (Length = 2**N)
    min_energy : float, default 0.2
        min_energy : percentage of energy a sampling rate must represent at least, to be detected. Lower values
        result in higher sensibillity - but as well increas detection rate of mix products. Default proofed to be
        stable.
    max_freqs : int, default 10
        Maximum number of frequencies collected from the index. Mainly a value to prevent the frequency
        collection loop from collecting noise and running endlessly.
    bins : {None, List[float]} : default None

    Returns
    -------
        freq : {None, str}
            Either the sampling rate that was detected in the sample index (if uniform). Or
            the greates common rate of all the sampling rates detected. Equals `None` if
            detection failed and `"empty"`, if input index was empty.
        freqs : List[str]
            List of detected sampling rates.

    """
    index_n = index.to_numpy(float)
    if index.empty:
        return "empty", []

    index_n = (index_n - index_n[0]) * 10 ** (-9 + delta_precision)
    delta = np.zeros(int(index_n[-1]) + 1)
    delta[index_n.astype(int)] = 1
    if optimize:
        delta_f = np.abs(fft.rfft(delta, fft.next_fast_len(len(delta))))
    else:
        delta_f = np.abs(fft.rfft(delta))

    len_f = len(delta_f) * 2
    min_energy = delta_f[0] * min_energy
    # calc/assign low/high freq cut offs (makes life easier):
    min_rate_i = int(
        len_f / (pd.Timedelta(min_rate).total_seconds() * (10**delta_precision))
    )
    delta_f[:min_rate_i] = 0
    max_rate_i = int(
        len_f / (pd.Timedelta(max_rate).total_seconds() * (10**delta_precision))
    )
    hf_cutoff = min(max_rate_i, len_f // 2)
    delta_f[hf_cutoff:] = 0
    delta_f[delta_f < min_energy] = 0

    # find frequencies present:
    freqs = []
    f_i = np.argmax(delta_f)
    while (f_i > 0) & (len(freqs) < max_freqs):
        f = (len_f / f_i) / (60 * 10 ** (delta_precision))
        freqs.append(f)
        for i in range(1, hf_cutoff // f_i + 1):
            delta_f[(i * f_i) - min_rate_i : (i * f_i) + min_rate_i] = 0
        f_i = np.argmax(delta_f)

    if len(freqs) == 0:
        return None, []

    if bins is None:
        r = range(0, int(pd.Timedelta(min_rate).total_seconds() / 60))
        bins = [0, 0.1, 0.2, 0.3, 0.4] + [i + 0.5 for i in r]

    f_hist, bins = np.histogram(freqs, bins=bins)
    freqs = np.ceil(bins[:-1][f_hist >= 1])
    gcd_freq = np.gcd.reduce((10 * freqs).astype(int)) / 10

    return str(int(gcd_freq)) + "min", [str(int(i)) + "min" for i in freqs]


def detectDeviants(
    data,
    metric,
    norm_spread,
    norm_frac,
    linkage_method="single",
    population="variables",
):
    """
    Helper function for carrying out the repeatedly upcoming task,
    of detecting variables a group of variables.

    "Normality" is determined in terms of a maximum spreading distance, that members of a normal group must not exceed
    in respect to a certain metric and linkage method.

    In addition, only a group is considered "normal" if it contains more then `frac` percent of the
    variables in "fields".

    Parameters
    ----------
    data : {pandas.DataFrame, DictOfSeries}
        Input data
    metric : Callable[[numpy.array, numpy.array], float]
        A metric function that for calculating the dissimilarity between 2 variables.
    norm_spread : float
        A threshold denoting the distance, members of the "normal" group must not exceed to each other (in terms of the
        metric passed) to qualify their group as the "normal" group.
    norm_frac : float, default 0.5
        Has to be in [0,1]. Determines the minimum percentage of variables or samples,
        the "normal" group has to comprise to be the normal group actually.
    linkage_method : {"single", "complete", "average", "weighted", "centroid", "median", "ward"}, default "single"
        The linkage method used for hierarchical (agglomerative) clustering of the variables.
    population : {"variables", "samples"}
        Wheather to relate the minimum percentage of values needed to form a normal group, to the total number of
        variables or to the total number of samples.

    Returns
    -------
    deviants : list
        A list containing the column positions of deviant variables in the input

    """
    var_num = len(data.columns)
    if var_num <= 1:
        return []
    dist_mat = np.zeros((var_num, var_num))
    combs = list(itertools.combinations(range(0, var_num), 2))
    for i, j in combs:
        d_i = data[data.columns[i]]
        d_j = data[data.columns[j]]
        dist = metric(d_i.values, d_j.values)
        dist_mat[i, j] = dist

    condensed = np.abs(dist_mat[tuple(zip(*combs))])
    Z = linkage(condensed, method=linkage_method)
    cluster = fcluster(Z, norm_spread, criterion="distance")
    if population == "variables":
        counts = collections.Counter(cluster)
        pop_num = var_num
    elif population == "samples":
        counts = {cluster[j]: 0 for j in range(0, var_num)}
        for c in range(var_num):
            field = data.columns[c]
            counts[cluster[c]] += data[field].dropna().shape[0]
        pop_num = np.sum(list(counts.values()))
    else:
        raise ValueError(
            "Not a valid normality criteria keyword passed. "
            "Pass either 'variables' or 'population'."
        )
    norm_cluster = -1

    for item in counts.items():
        if item[1] > norm_frac * pop_num:
            norm_cluster = item[0]
            break

    if norm_cluster == -1 or counts[norm_cluster] == pop_num:
        return []
    else:
        return [i for i, x in enumerate(cluster) if x != norm_cluster]


def getFreqDelta(index: pd.Index) -> None | pd.Timedelta:
    """
    Function checks if the passed index is regularly sampled.

    If yes, the according timedelta value is returned,

    If no, ``None`` is returned.

    (``None`` will also be returned for pd.RangeIndex type.)

    """
    delta = getattr(index, "freq", None)
    if delta is None and not index.empty:
        i = pd.date_range(index[0], index[-1], len(index))
        if i.equals(index):
            return i[1] - i[0]
    return delta


def getApply(in_obj, apply_obj, attr_access="__name__", attr_or="apply") -> pd.Series:
    """
    For the repeating task of applying build in (accelerated) methods/funcs (`apply_obj`),
    of rolling/resampling - like objects (`in_obj`) ,
    if those build-ins are available, or pass the method/func to the objects apply-like method, otherwise.

    """
    try:
        out = getattr(in_obj, getattr(apply_obj, attr_access))()
    except AttributeError:
        try:
            # let's try to run it somewhat optimized
            out = getattr(in_obj, attr_or)(apply_obj, raw=True)
        except:
            # did't work out, fallback
            out = getattr(in_obj, attr_or)(apply_obj)

    return out


def statPass(
    datcol: pd.Series,
    stat: Callable[[np.ndarray, pd.Series], float],
    winsz: pd.Timedelta,
    thresh: float,
    comparator: Callable[[CompT, CompT], bool],
    sub_winsz: pd.Timedelta | None = None,
    sub_thresh: float | None = None,
    min_periods: int | None = None,
) -> pd.Series:
    """
    Check `datcol`, if it contains chunks of length `window`, exceeding `thresh` with
    regard to `func` and `comparator`:

    (check, if: `comparator`(func`(*chunk*), `thresh`)

    If yes, subsequently check, if all (maybe overlapping) *sub-chunks* of *chunk*, with length `sub_window`,
    satisfy, `comparator`(`func`(*sub_chunk*), `sub_thresh`)

    returns boolean series with same index as input series
    """
    stat_parent = datcol.rolling(winsz, min_periods=min_periods)
    stat_parent = getApply(stat_parent, stat)
    exceeds = comparator(stat_parent, thresh)
    if sub_winsz:
        stat_sub = datcol.rolling(sub_winsz)
        stat_sub = getApply(stat_sub, stat)
        min_stat = stat_sub.rolling(winsz - sub_winsz, closed="both").min()
        exceeding_sub = comparator(min_stat, sub_thresh)
        exceeds = exceeding_sub & exceeds

    to_set = pd.Series(False, index=exceeds.index)
    exceeds_df = pd.DataFrame(
        {"g": exceeds.diff().cumsum().values, "ex_val": exceeds.values},
        index=exceeds.index,
    )
    for _, group in exceeds_df.groupby(by="g"):
        if group["ex_val"].iloc[0]:
            # dt-slices include both bounds, so we subtract 1ns
            start = group.index[0] - (winsz - pd.Timedelta("1ns"))
            end = group.index[-1]
            to_set[start:end] = True

    return to_set


def filterKwargs(
    kwargs: dict,
    reserved: Collection,
    inplace: bool = True,
    warn: bool = True,
    msg: str = "",
    stacklevel: int = 3,
) -> dict:
    """
    Filter kwargs (or any dict) by a list of reserved keys.

    Parameters
    ----------
    kwargs : dict
        The dict to filter.

    reserved : list-like
        A list of reserved keywords.

    inplace : bool, default False
        If `False` a copy is returned, otherwise the modified original kwargs.

    warn : bool, default True
        Throw a `RuntimeWarning` with the following text:

    msg : str, default ""
        A text to append to the warnings message.

    stacklevel : int, default 3
        The stacklevel this warning will refer to.
            - `2` : warn at the location this function is called
            - `3` : warn for the function that calls this function (default)

    Returns
    -------
    kwargs: dict
        the modified kwargs or a copy
    """
    if not inplace:
        kwargs = kwargs.copy()
    for key in reserved:
        if warn and key in kwargs:
            warnings.warn(
                f"The keyword {repr(key)} is reserved and will be ignored {msg}",
                RuntimeWarning,
                stacklevel=stacklevel,
            )
        kwargs.pop(key, None)
    return kwargs


from saqc import FILTER_ALL, UNFLAGGED

A = TypeVar("A", np.ndarray, pd.Series)


def isflagged(flagscol: A, thresh: float) -> A:
    """
    Return a mask of flags accordingly to `thresh`. Return type is same as flags.
    """
    if not isinstance(thresh, (float, int)):
        raise TypeError(f"thresh must be of type float, not {repr(type(thresh))}")

    if thresh == FILTER_ALL:
        return flagscol > UNFLAGGED

    return flagscol >= thresh


def isunflagged(flagscol: A, thresh: float) -> A:
    return ~isflagged(flagscol, thresh)


def getUnionIndex(obj, default: pd.DatetimeIndex | None = None):
    assert hasattr(obj, "columns")
    if default is None:
        default = pd.DatetimeIndex([])
    indexes = [obj[k].index for k in obj.columns]
    if indexes:
        return functools.reduce(pd.Index.union, indexes).sort_values()
    return default


def getSharedIndex(obj, default: pd.DatetimeIndex | None = None):
    assert hasattr(obj, "columns")
    if default is None:
        default = pd.DatetimeIndex([])
    indexes = [obj[k].index for k in obj.columns]
    if indexes:
        return functools.reduce(pd.Index.intersection, indexes).sort_values()
    return default


def isAllBoolean(obj: Any):
    if not hasattr(obj, "columns"):
        return pd.api.types.is_bool_dtype(obj)
    for c in obj.columns:
        if not pd.api.types.is_bool_dtype(obj[c]):
            return False
    return True



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-
from __future__ import annotations

import abc
from typing import Any, Dict, TypeVar, Union

import numpy as np
import pandas as pd
from typing_extensions import Protocol

__all__ = [
    "T",
    "ArrayLike",
    "CurveFitter",
    "ExternalFlag",
    "OptionalNone",
]

T = TypeVar("T")
ArrayLike = TypeVar("ArrayLike", np.ndarray, pd.Series, pd.DataFrame)

ExternalFlag = Union[str, float, int]


# needed for deeper type hinting magic
class CurveFitter(Protocol):
    def __call__(self, data: np.ndarray, *params: float) -> np.ndarray:
        ...  # pragma: no cover


class Comparable(Protocol):
    @abc.abstractmethod
    def __gt__(self: CompT, other: CompT) -> bool:
        pass


CompT = TypeVar("CompT", bound=Comparable)


class OptionalNone:
    pass



#!/usr/bin/env python
# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import functools
import math

import numpy as np
import pandas as pd


def removeRollingRamps(
    data: pd.Series,
    window: int | str | pd.Timedelta,
    center: bool = False,
    inplace=False,
) -> pd.Series:
    """
    Set data to `NaN`, where the window of prior rolling function was
    not fully shifted into the data.

    Pandas rolling implementation shift a window over the data from left to right.
    The window starts on the far left, outside the data [d], and the window result
    is written [D] on the right edge of the window.

        data:          [d d d d d d ... d d d]
        window(0): [- - D] ==>
        window(1):   [- d D] ==>
        window(2):     [d d D] ==>
           ...                ...
        window(n-1):              ==> [d d D]
        window(n):                  ==> [d d D]

    This results in some windows at the start, where the window is not shifted in
    completely (first two windows in the example above). This might be unexpected or
    undesired for some calculations/algorithms that assume the window is 'filled
    with data'.
    For example: A constant checker that wants to check if a whole day of data is
    within a specific small range, will define a window of `1day`. The data is
    sampled on a 5-minutes base. So the fist window will have one value, the
    second two values, the third three vales etc. until the window is completely
    moved into the data. This might result in the checker marking the beginning of
    the data as constant because it only compares few values. Setting
    ``min_periods`` might help, but for time-based windows which roll over a
    non-regular grid (the time interval between the datapoints is not regular) it
    is not possible to define a value like a 'a full window', because one can not
    know how many (irregular spaced) data-points will be in the window.
    Therefore, this function sets windows, that where not fully shifted into the data
    to ``np.nan``.
    Pass the same arguments for ``window``, ``closed`` and ``center`` as were passed to
    rolling.

    Parameters
    ----------
    data :
        Data resulted from a rolling operation.
    window :
        Pass the same value as was passed to the prior called pd.rolling.
    center :
        Pass the same value as was passed to the prior called pd.rolling.
    closed :
        Pass the same value as was passed to the prior called pd.rolling.
    inplace :
        Return a modified copy if False, otherwise return the modified
        passed in data.

    Returns
    -------
    data: pd.Series
        updated original data or copy

    Examples
    --------

    >>> s = pd.Series(1, index=pd.date_range('2000-02-02', freq='1d', periods=4), dtype=int)
    >>> s
    Out[6]:
    2000-01-01    1
    2000-01-02    1
    2000-01-03    1
    2000-01-04    1
    Freq: D, dtype: int64

    If we roll with a window of 2 days, the result should have NaNs where the window
    is not fully shifted in the data. For the timestamp `2000-02-03 00:00:00` the start
    of the window would be `2000-02-01 00:00:00` which lies outside the data and
    therefore the result is set to NaN.

    >>> dat = s.rolling('2d').sum()
    >>> res = removeRollingRamps(dat, '2d')
    >>> pd.DataFrame(dict(dat=dat, res=res))
                dat  res
    2000-02-02  1.0  NaN
    2000-02-03  2.0  NaN
    2000-02-04  2.0  2.0
    2000-02-05  2.0  2.0

    >>> dat = s.rolling('2d10min').sum()
    >>> res = removeRollingRamps(dat, '2d10min')
    >>> pd.DataFrame(dict(dat=dat, res=res))
                dat  res
    2000-02-02  1.0  NaN
    2000-02-03  2.0  NaN
    2000-02-04  3.0  NaN
    2000-02-05  3.0  3.0

    together with the center keyword the NaNs will appear on both sides of the data
    series, because the window is shifted into the data and also shifts out of it.

    >>> dat = s.rolling('2d', center=True).sum()
    >>> res = removeRollingRamps(dat, '2d', center=True)
    >>> pd.DataFrame(dict(dat=dat, res=res))
                dat  res
    2000-02-02  2.0  NaN
    2000-02-03  2.0  2.0
    2000-02-04  2.0  2.0
    2000-02-05  1.0  NaN
    """
    if not inplace:
        data = data.copy()

    # this ensures index[0] and index[-1]
    if data.empty:
        return data

    if isinstance(window, int):
        window = max(window - 1, 0)
        if center:
            left = math.ceil(window / 2)
            right = math.floor(window / 2)
            data.iloc[:left] = np.nan
            data.iloc[-right:] = np.nan
        else:
            data.iloc[:window] = np.nan
    else:
        window = pd.Timedelta(window)
        if center:
            window //= 2
            data.loc[data.index < data.index[0] + window] = np.nan
            data.loc[data.index > data.index[-1] - window] = np.nan
        else:
            data.loc[data.index < data.index[0] + window] = np.nan
    return data



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-

import pandas as pd
import pytest

from saqc import BAD, UNFLAGGED, SaQC
from saqc.constants import FILTER_NONE
from saqc.core import DictOfSeries, Flags
from tests.common import initData


@pytest.fixture
def data():
    return initData()


def test_emptyData():
    # test that things do not break with empty data sets
    saqc = SaQC(data=pd.DataFrame({"x": [], "y": []}))

    saqc.flagGeneric("x", func=lambda x: x < 0)
    assert saqc.data.empty
    assert saqc.flags.empty

    saqc = saqc.processGeneric(field="x", target="y", func=lambda x: x + 2)
    assert saqc.data.empty
    assert saqc.flags.empty


@pytest.mark.parametrize(
    "targets, func",
    [
        (["tmp"], lambda x, y: pd.Series(True, index=x.index.union(y.index))),
        (
            ["tmp1", "tmp2"],
            lambda x, y: [pd.Series(True, index=x.index.union(y.index))] * 2,
        ),
    ],
)
def test_writeTargetFlagGeneric(data, targets, func):
    saqc = SaQC(data=data)
    saqc = saqc.flagGeneric(field=data.columns, target=targets, func=func, flag=BAD)
    for target in targets:
        assert saqc._flags.history[target].hist.iloc[0].tolist() == [BAD]


@pytest.mark.parametrize(
    "fields, func",
    [
        (["var1"], lambda x: pd.Series(True, index=x.index)),
        (
            ["var1", "var2"],
            lambda x, y: [pd.Series(True, index=x.index.union(y.index))] * 2,
        ),
    ],
)
def test_overwriteFieldFlagGeneric(data, fields, func):
    flag = 12

    saqc = SaQC(
        data=data.copy(),
        flags=Flags(
            {
                k: pd.Series(data[k] % 2, index=data[k].index).replace(
                    {0: UNFLAGGED, 1: 127}
                )
                for k in data.columns
            }
        ),
    )

    res = saqc.flagGeneric(field=fields, func=func, flag=flag, dfilter=FILTER_NONE)
    for field in fields:
        histcol1 = res._flags.history[field].hist[1]
        assert (histcol1 == flag).all()
        assert (data[field] == res.data[field]).all(axis=None)
        assert res._flags.history[field].meta[0] == {}


@pytest.mark.parametrize(
    "data, targets, func, expected_data",
    [
        (
            DictOfSeries(dict(a=pd.Series([1.0, 2.0]), b=pd.Series([10.0, 20.0]))),
            ["t1"],
            lambda x, y: x + y,
            DictOfSeries(
                dict(
                    a=pd.Series([1.0, 2.0]),
                    b=pd.Series([10.0, 20.0]),
                    t1=pd.Series([11.0, 22.0]),
                )
            ),
        ),
        (
            DictOfSeries(dict(a=pd.Series([1.0, 2.0]), b=pd.Series([10.0, 20.0]))),
            ["t1", "t2"],
            lambda x, y: (x + y, y * 2),
            DictOfSeries(
                dict(
                    a=pd.Series([1.0, 2.0]),
                    b=pd.Series([10.0, 20.0]),
                    t1=pd.Series([11.0, 22.0]),
                    t2=pd.Series([20.0, 40.0]),
                )
            ),
        ),
    ],
)
def test_writeTargetProcGeneric(data, targets, func, expected_data):
    fields = data.columns.tolist()
    dfilter = 128

    saqc = SaQC(
        data=data,
        flags=Flags({k: pd.Series(127.0, index=data[k].index) for k in data.columns}),
    )
    res = saqc.processGeneric(
        field=fields,
        target=targets,
        func=func,
        dfilter=dfilter,
        label="generic",
    )
    assert expected_data == res.data
    # check that new histories where created
    for target in targets:
        assert res._flags.history[target].hist.iloc[0].isna().all()


@pytest.mark.parametrize(
    "data, fields, func, expected_data",
    [
        (
            DictOfSeries(dict(a=pd.Series([1.0, 2.0]), b=pd.Series([10.0, 20.0]))),
            ["a"],
            lambda x: x * 2,
            DictOfSeries(dict(a=pd.Series([2.0, 4.0]), b=pd.Series([10.0, 20.0]))),
        ),
        (
            DictOfSeries(dict(a=pd.Series([1.0, 2.0]), b=pd.Series([10.0, 20.0]))),
            ["a", "b"],
            lambda x, y: (x + y, y * 2),
            DictOfSeries(dict(a=pd.Series([11.0, 22.0]), b=pd.Series([20.0, 40.0]))),
        ),
    ],
)
def test_overwriteFieldProcGeneric(data, fields, func, expected_data):
    dfilter = 128

    saqc = SaQC(
        data=data,
        flags=Flags({k: pd.Series(127.0, index=data[k].index) for k in data.columns}),
    )

    res = saqc.processGeneric(field=fields, func=func, dfilter=dfilter, label="generic")
    assert expected_data == res.data
    # check that the histories got appended
    for field in fields:
        assert (res._flags.history[field].hist[0] == 127.0).all()
        assert res._flags.history[field].hist[1].isna().all()
        assert res._flags.history[field].meta[0] == {}


def test_label():
    dat = pd.DataFrame(
        {"data1": [1, 1, 5, 2, 1], "data2": [1, 1, 2, 3, 4], "data3": [1, 1, 2, 3, 4]},
        index=pd.date_range("2000", "2005", periods=5),
    )

    qc = SaQC(dat)
    qc = qc.flagRange("data1", max=4, label="out of range")
    qc = qc.flagRange("data1", max=0, label="out of range2")
    qc = qc.flagGeneric(
        ["data1", "data3"],
        target="data2",
        func=lambda x, y: isflagged(x, "out of range") | isflagged(y),  # noqa
    )
    assert list((qc.flags["data2"] > 0).values) == [False, False, True, False, False]


@pytest.mark.parametrize(
    "kwargs, got, expected",
    [
        (
            {
                "lower": 0,
            },
            [-9, -2, 1, 2, 9],
            [0, 0, 1, 2, 9],
        ),
        ({"upper": 3}, [-9, -2, 1, 2, 9], [-9, -2, 1, 2, 3]),
        ({"lower": -6, "upper": 3}, [-9, -2, 1, 2, 9], [-6, -2, 1, 2, 3]),
    ],
)
def test_processGenericClip(kwargs, got, expected):
    field = "data"
    got = pd.DataFrame(
        got, columns=[field], index=pd.date_range("2020-06-30", periods=len(got))
    )
    expected = pd.DataFrame(
        expected,
        columns=[field],
        index=pd.date_range("2020-06-30", periods=len(expected)),
    )
    qc = SaQC(got).processGeneric(field, func=lambda x: clip(x, **kwargs))
    assert (qc._data[field] == expected[field]).all()



#!/usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
The module comprises flag value constants in use throughout saqc.

Flagging Constants
------------------
* :py:const:`~saqc.constants.UNFLAGGED`: indicates that no flag has been assigned yet
* :py:const:`~saqc.constants.GOOD`: the lowest possible flag value larget than :py:const:`~saqc.constants.UNFLAGGED`
* :py:const:`~saqc.constants.DOUBFUL`: a modest flag value, usually indicating some sort of suspiciousness
* :py:const:`~saqc.constants.BAD`: the highest flag value available

The flagging constants are ordered (from "worse" to "best") as:

:py:const:`~saqc.constants.BAD` > :py:const:`~saqc.constants.DOUBTFUL` > :py:const:`~saqc.constants.GOOD` > :py:const:`~saqc.constants.UNFLAGGED`

Filtering Constants
-------------------
* :py:const:`~saqc.constants.FILTER_ALL`: mask/filter all flagged data
* :py:const:`~saqc.constants.FILTER_NONE`: mask/filter no data
"""

__all__ = [
    "UNFLAGGED",
    "DOUBTFUL",
    "BAD",
    "GOOD",
    "FILTER_ALL",
    "FILTER_NONE",
]


import numpy as np

# ----------------------------------------------------------------------
# global flag constants
# ----------------------------------------------------------------------

UNFLAGGED = -np.inf
GOOD = 0
DOUBTFUL = 25.0
BAD = 255.0

# ----------------------------------------------------------------------
# global dfilter constants
# ----------------------------------------------------------------------

FILTER_ALL = -np.inf
FILTER_NONE = np.inf



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Callable, Tuple, Union

import numpy as np
import pandas as pd
from typing_extensions import Literal

from saqc import UNFLAGGED
from saqc.core import register
from saqc.core.history import History
from saqc.lib.tools import isflagged
from saqc.lib.ts_operators import interpolateNANs, shift2Freq

if TYPE_CHECKING:
    from saqc import SaQC


# TODO: remove, when `interpolateIndex` and `interpolateInvalid are removed`
INTERPOLATION_METHODS = Literal[
    "linear",
    "time",
    "nearest",
    "zero",
    "slinear",
    "quadratic",
    "cubic",
    "spline",
    "barycentric",
    "polynomial",
    "krogh",
    "piecewise_polynomial",
    "spline",
    "pchip",
    "akima",
]


def _resampleOverlapping(data: pd.Series, freq: str, fill_value):
    """TODO: docstring needed"""
    dtype = data.dtype
    end = data.index[-1].ceil(freq)
    data = data.resample(freq).max()
    data = data.combine(data.shift(1, fill_value=fill_value), max)
    if end not in data:
        data.loc[end] = fill_value
    return data.fillna(fill_value).astype(dtype)


class InterpolationMixin:
    @register(
        mask=["field"],
        demask=["field"],
        squeeze=[],  # func handles history by itself
    )
    def interpolateByRolling(
        self: "SaQC",
        field: str,
        window: Union[str, int],
        func: Callable[[pd.Series], float] = np.median,
        center: bool = True,
        min_periods: int = 0,
        flag: float = UNFLAGGED,
        **kwargs,
    ) -> "SaQC":
        """
        Interpolates nan-values in the data by assigning them the aggregation result of the window surrounding them.

        Parameters
        ----------
        window :
            The size of the window, the aggregation is computed from. An integer define the number of periods to be used,
            an string is interpreted as an offset. ( see `pandas.rolling` for more information).
            Integer windows may result in screwed aggregations if called on none-harmonized or irregular data.

        func : default median
            The function used for aggregation.

        center :
            Center the window around the value. Can only be used with integer windows, otherwise it is silently ignored.

        min_periods :
            Minimum number of valid (not np.nan) values that have to be available in a window for its aggregation to be
            computed.
        """
        datcol = self._data[field]
        roller = datcol.rolling(window=window, center=center, min_periods=min_periods)
        try:
            func_name = func.__name__
            if func_name[:3] == "nan":
                func_name = func_name[3:]
            rolled = getattr(roller, func_name)()
        except AttributeError:
            rolled = roller.apply(func)

        na_mask = datcol.isna()
        interpolated = na_mask & rolled.notna()
        datcol[na_mask] = rolled[na_mask]
        self._data[field] = datcol

        flagcol = pd.Series(np.nan, index=self._flags[field].index)
        flagcol.loc[interpolated] = np.nan if flag is None else flag

        # todo kwargs must have all passed args except data,field,flags
        meta = {
            "func": "interpolateByRolling",
            "args": (field,),
            "kwargs": {
                "window": window,
                "func": func,
                "center": center,
                "min_periods": min_periods,
                "flag": flag,
                **kwargs,
            },
        }
        self._flags.history[field].append(flagcol, meta)

        return self

    @register(
        mask=["field"],
        demask=[],
        squeeze=[],  # func handles history by itself
    )
    def interpolate(
        self: "SaQC",
        field: str,
        method: INTERPOLATION_METHODS = "time",
        order: int = 2,
        limit: int | str | None = None,
        extrapolate: Literal["forward", "backward", "both"] | None = None,
        flag: float = UNFLAGGED,
        **kwargs,
    ) -> "SaQC":
        """
        Fill NaN and flagged values using an interpolation method.

        Parameters
        ----------
        method :
            Interpolation technique to use. One of:

            * ‘linear’: Ignore the index and treat the values as equally spaced.
            * ‘time’: Works on daily and higher resolution data to interpolate given length of interval.
            * ‘index’, ‘values’: Use the actual numerical values of the index.
            * ‘pad’: Fill in NaNs using existing values.
            * ‘nearest’, ‘zero’, ‘slinear’, ‘quadratic’, ‘cubic’, ‘spline’, ‘barycentric’, ‘polynomial’:
                 Passed to scipy.interpolate.interp1d. These methods use the numerical values of the index.
                 Both ‘polynomial’ and ‘spline’ require that you also specify an order (int), e.g.
                 ``qc.interpolate(method='polynomial', order=5)``.
            * ‘krogh’, ‘spline’, ‘pchip’, ‘akima’, ‘cubicspline’:
                 Wrappers around the SciPy interpolation methods of similar names.
            * ‘from_derivatives’: Refers to scipy.interpolate.BPoly.from_derivatives

        order :
            Order of the interpolation method, ignored if not supported by the chosen ``method``

        limit :
            Maximum number of missing values to interpolate. Only gaps smaller than ``limit`` will be filled.
            The gap size can be given as a number of values (integer) or a temporal extensions (offset string).
            With ``None``, all missing values will be interpolated.

        extrapolate :
            Use parameter to perform extrapolation instead of interpolation onto the trailing and/or leading chunks of
            NaN values in data series.

            * 'None' (default) - perform interpolation
            * 'forward'/'backward' - perform forward/backward extrapolation
            * 'both' - perform forward and backward extrapolation

        Examples
        --------
        See some examples of the keyword interplay below:

        Lets generate some dummy data:

        .. doctest:: interpolate

           >>> data = pd.DataFrame({'data':np.array([np.nan, 0, np.nan, np.nan, np.nan, 4, 5, np.nan, np.nan, 8, 9, np.nan, np.nan])}, index=pd.date_range('2000',freq='1H', periods=13))
           >>> data
                                data
           2000-01-01 00:00:00   NaN
           2000-01-01 01:00:00   0.0
           2000-01-01 02:00:00   NaN
           2000-01-01 03:00:00   NaN
           2000-01-01 04:00:00   NaN
           2000-01-01 05:00:00   4.0
           2000-01-01 06:00:00   5.0
           2000-01-01 07:00:00   NaN
           2000-01-01 08:00:00   NaN
           2000-01-01 09:00:00   8.0
           2000-01-01 10:00:00   9.0
           2000-01-01 11:00:00   NaN
           2000-01-01 12:00:00   NaN

        Use :py:meth:`~saqc.SaQC.interpolate` to do linear interpolation of up to 2 consecutive missing values:

        .. doctest:: interpolate

           >>> qc = saqc.SaQC(data)
           >>> qc = qc.interpolate("data", limit=3, method='time')
           >>> qc.data # doctest:+NORMALIZE_WHITESPACE
                               data |
           ======================== |
           2000-01-01 00:00:00  NaN |
           2000-01-01 01:00:00  0.0 |
           2000-01-01 02:00:00  NaN |
           2000-01-01 03:00:00  NaN |
           2000-01-01 04:00:00  NaN |
           2000-01-01 05:00:00  4.0 |
           2000-01-01 06:00:00  5.0 |
           2000-01-01 07:00:00  6.0 |
           2000-01-01 08:00:00  7.0 |
           2000-01-01 09:00:00  8.0 |
           2000-01-01 10:00:00  9.0 |
           2000-01-01 11:00:00  NaN |
           2000-01-01 12:00:00  NaN |
           <BLANKLINE>


        Use :py:meth:`~saqc.SaQC.interpolate` to do linear extrapolaiton of up to 1 consecutive missing values:

        .. doctest:: interpolate

           >>> qc = saqc.SaQC(data)
           >>> qc = qc.interpolate("data", limit=2, method='time', extrapolate='both')
           >>> qc.data # doctest:+NORMALIZE_WHITESPACE
                               data |
           ======================== |
           2000-01-01 00:00:00  0.0 |
           2000-01-01 01:00:00  0.0 |
           2000-01-01 02:00:00  NaN |
           2000-01-01 03:00:00  NaN |
           2000-01-01 04:00:00  NaN |
           2000-01-01 05:00:00  4.0 |
           2000-01-01 06:00:00  5.0 |
           2000-01-01 07:00:00  NaN |
           2000-01-01 08:00:00  NaN |
           2000-01-01 09:00:00  8.0 |
           2000-01-01 10:00:00  9.0 |
           2000-01-01 11:00:00  NaN |
           2000-01-01 12:00:00  NaN |
           <BLANKLINE>
        """

        if "freq" in kwargs:
            # the old interpolate version
            warnings.warn(
                f"""
                The method `intepolate` is deprecated and will be removed in version 3.0 of saqc.
                To achieve the same behaviour please use:
                `qc.align(field={field}, freq={kwargs["freq"]}, method={method}, order={order}, flag={flag})`
                """,
                DeprecationWarning,
            )
            return self.align(
                field=field,
                freq=kwargs.pop("freq", method),
                method=method,
                order=order,
                flag=flag,
                **kwargs,
            )

        inter_data = interpolateNANs(
            self._data[field],
            method,
            order=order,
            gap_limit=limit,
            extrapolate=extrapolate,
        )

        interpolated = self._data[field].isna() & inter_data.notna()
        self._data[field] = inter_data
        new_col = pd.Series(np.nan, index=self._flags[field].index)
        new_col.loc[interpolated] = np.nan if flag is None else flag

        # todo kwargs must have all passed args except data,field,flags
        self._flags.history[field].append(
            new_col, {"func": "interpolateInvalid", "args": (), "kwargs": kwargs}
        )

        return self

    @register(mask=["field"], demask=[], squeeze=[])
    def align(
        self: "SaQC",
        field: str,
        freq: str,
        method: INTERPOLATION_METHODS = "time",
        order: int = 2,
        extrapolate: Literal["forward", "backward", "both"] | None = None,
        overwrite: bool = False,
        **kwargs,
    ) -> "SaQC":
        """
        Convert time series to specified frequency. Values affected by frequency
        changes will be inteprolated using the given method.

        Parameters
        ----------
        freq :
            Target frequency.

        method :
            Interpolation technique to use. One of:

            * ``'nshift'``: shift grid points to the nearest time stamp in the range = +/- 0.5 * ``freq``
            * ``'bshift'``: shift grid points to the first succeeding time stamp (if any)
            * ``'fshift'``: shift grid points to the last preceeding time stamp (if any)
            * ``'linear'``: Ignore the index and treat the values as equally spaced.
            * ``'time'``, ``'index'``, 'values': Use the actual numerical values of the index.
            * ``'pad'``: Fill in NaNs using existing values.
            * ``'nearest'``, ``'zero'``, ``'slinear'``, ``'quadratic'``, ``'cubic'``, ``'spline'``, ``'barycentric'``, ``'polynomial'``:
              Passed to ``scipy.interpolate.interp1d``. These methods use the numerical values of the index. Both ``'polynomial'`` and
              ``'spline'`` require that you also specify an ``order``, e.g. ``qc.interpolate(method='polynomial', order=5)``.
            * ``'krogh'``, ``'spline'``, ``'pchip'``, ``'akima'``, ``'cubicspline'``:
              Wrappers around the SciPy interpolation methods of similar names.
            * ``'from_derivatives'``: Refers to ``scipy.interpolate.BPoly.from_derivatives``

        order :
            Order of the interpolation method, ignored if not supported by the chosen ``method``

        extrapolate :
            Use parameter to perform extrapolation instead of interpolation onto the trailing and/or leading chunks of
            NaN values in data series.

            * ``None`` (default) - perform interpolation
            * ``'forward'``/``'backward'`` - perform forward/backward extrapolation
            * ``'both'`` - perform forward and backward extrapolation

        overwrite :
           If set to True, existing flags will be cleared
        """

        # TODO:
        # - should we keep `extrapolate`

        if self._data[field].empty:
            return self

        if method in ("fshift", "bshift", "nshift"):
            datacol, history = _shift(
                saqc=self, field=field, freq=freq, method=method, **kwargs
            )
        else:
            datacol, history = _interpolate(
                saqc=self,
                field=field,
                freq=freq,
                method=method,
                order=order,
                extrapolate=extrapolate,
                dfilter=kwargs["dfilter"],
            )

        meta = {
            "func": "align",
            "args": (field,),
            "kwargs": {
                "freq": freq,
                "method": method,
                "order": order,
                "extrapolate": extrapolate,
                **kwargs,
            },
        }

        flagcol = pd.Series(UNFLAGGED if overwrite else np.nan, index=history.index)
        history.append(flagcol, meta)

        self._data[field] = datacol
        self._flags.history[field] = history

        return self

    ### Deprecated functions

    @register(mask=["field"], demask=[], squeeze=[])
    def interpolateIndex(
        self: "SaQC",
        field: str,
        freq: str,
        method: INTERPOLATION_METHODS,
        order: int = 2,
        limit: int | None = 2,
        extrapolate: Literal["forward", "backward", "both"] = None,
        **kwargs,
    ) -> "SaQC":
        """
        Function to interpolate the data at regular (äquidistant) timestamps (or Grid points).

        Parameters
        ----------
        freq :
            An Offset String, interpreted as the frequency of
            the grid you want to interpolate your data to.

        method :
            The interpolation method you want to apply.

        order :
            If your selected interpolation method can be performed at different 'orders' - here you pass the desired
            order.

        limit :
            Upper limit of missing index values (with respect to ``freq``) to fill. The limit can either be expressed
            as the number of consecutive missing values (integer) or temporal extension of the gaps to be filled
            (Offset String).
            If ``None`` is passed, no limit is set.

        extraplate :
            Use parameter to perform extrapolation instead of interpolation onto the trailing and/or leading chunks of
            NaN values in data series.

            * ``None`` (default) - perform interpolation
            * ``'forward'``/``'backward'`` - perform forward/backward extrapolation
            * ``'both'`` - perform forward and backward extrapolation
        """

        msg = """
        The method `interpolateIndex` is deprecated and will be removed in verion 3.0 of saqc.
        To achieve the same behavior use:
        """
        call = "qc.align(field={field}, freq={freq}, method={method}, order={order}, extrapolate={extrapolate})"
        if limit != 2:
            call = f"{call}.interpolate(field={field}, method={method}, order={order}, limit={limit}, extrapolate={extrapolate})"

        warnings.warn(f"{msg}`{call}`", DeprecationWarning)
        out = self.align(
            field=field,
            freq=freq,
            method=method,
            order=order,
            extrapolate=extrapolate,
            **kwargs,
        )
        if limit != 2:
            out = out.interpolate(
                field=field,
                freq=freq,
                method=method,
                order=order,
                limit=limit,
                extrapolate=extrapolate,
                **kwargs,
            )
        return out

    @register(
        mask=["field"],
        demask=["field"],
        squeeze=[],  # func handles history by itself
    )
    def interpolateInvalid(
        self: "SaQC",
        field: str,
        method: INTERPOLATION_METHODS,
        order: int = 2,
        limit: int | None = None,
        extrapolate: Literal["forward", "backward", "both"] | None = None,
        flag: float = UNFLAGGED,
        **kwargs,
    ) -> "SaQC":
        warnings.warn(
            f"""
            The method `intepolateInvalid` is deprecated and will be removed
            with version 3.0 of saqc. To achieve the same behavior, please use
            `qc.interpolate(
                field={field}, method={method}, order={order},
                limit={limit}, extrapolate={extrapolate}, flag={flag}
            )`
            """
        )

        return self.interpolate(
            field=field,
            method=method,
            order=order,
            limit=limit,
            extrapolate=extrapolate,
            flag=flag,
            **kwargs,
        )


def _shift(
    saqc: "SaQC",
    field: str,
    freq: str,
    method: Literal["fshift", "bshift", "nshift"] = "nshift",
    **kwargs,
) -> Tuple[pd.Series, History]:
    """
    Shift data points and flags to a regular frequency grid.

    Parameters
    ----------
    field :
        The fieldname of the column, holding the data-to-be-shifted.

    freq :
        Offset string. Sampling rate of the target frequency.

    method :
        Method to propagate values:

        * 'nshift' : shift grid points to the nearest time stamp in the range = +/- 0.5 * ``freq``
        * 'bshift' : shift grid points to the first succeeding time stamp (if any)
        * 'fshift' : shift grid points to the last preceeding time stamp (if any)

    freq_check :
        * ``None`` : do not validate the ``freq`` string.
        * 'check' : check ``freq`` against an frequency estimation, produces a warning in case of miss matches.
        * 'auto' : estimate frequency, `freq` is ignored.

    Returns
    -------
    saqc.SaQC
    """
    # TODO
    # - Do we need `freq_check`? If so could we move it to `align`?

    datcol = saqc._data[field]
    if datcol.empty:
        return saqc

    # do the shift
    datcol = shift2Freq(datcol, method, freq, fill_value=np.nan)

    # do the shift on the history
    kws = dict(method=method, freq=freq)

    history = saqc._flags.history[field].apply(
        index=datcol.index,
        func_handle_df=True,
        func=shift2Freq,
        func_kws={**kws, "fill_value": np.nan},
    )

    return datcol, history


def _interpolate(
    saqc: "SaQC",
    field: str,
    freq: str,
    method: str,
    order: int | None,
    dfilter: float,
    extrapolate: Literal["forward", "backward", "both"] | None = None,
) -> Tuple[pd.Series, History]:
    datcol = saqc._data[field].copy()

    start, end = datcol.index[0].floor(freq), datcol.index[-1].ceil(freq)
    grid_index = pd.date_range(start=start, end=end, freq=freq, name=datcol.index.name)

    flagged = isflagged(saqc._flags[field], dfilter)

    # drop all points that hold no relevant grid information
    datcol = datcol[~flagged].dropna()

    # account for annoying case of subsequent frequency aligned values,
    # that differ exactly by the margin of 2*freq
    gaps = datcol.index[1:] - datcol.index[:-1] == 2 * pd.Timedelta(freq)
    gaps = datcol.index[1:][gaps]
    gaps = gaps.intersection(grid_index).shift(-1, freq)

    # prepare grid interpolation:
    datcol = datcol.reindex(datcol.index.union(grid_index))

    # do the grid interpolation
    inter_data = interpolateNANs(
        data=datcol,
        method=method,
        order=order,
        gap_limit=2,
        extrapolate=extrapolate,
    )
    # override falsely interpolatet values:
    inter_data[gaps] = np.nan
    inter_data = inter_data[grid_index]

    history = saqc._flags.history[field].apply(
        index=inter_data.index,
        func=_resampleOverlapping,
        func_kws=dict(freq=freq, fill_value=np.nan),
    )
    return inter_data, history



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-

import copy

import numpy as np
import pandas as pd
import pytest

from saqc import BAD, FILTER_ALL, FILTER_NONE, UNFLAGGED, SaQC
from saqc.core import DictOfSeries, Flags, flagging, initFlagsLike, processing, register
from saqc.lib.types import OptionalNone
from tests.common import initData

OPTIONAL = [False, True]


@pytest.fixture
def data():
    return initData(3)


@pytest.fixture
def flags(data, optional):
    if optional:
        return None
    return DictOfSeries(initFlagsLike(data[data.columns[::2]]))


def test_errorHandling(data):
    @processing()
    def raisingFunc(saqc, field, **kwargs):
        raise TypeError

    var1 = data.columns[0]
    qc = SaQC(data)

    with pytest.raises(TypeError):
        qc.raisingFunc(var1)


@pytest.mark.parametrize("optional", OPTIONAL)
def test_dtypes(data, flags):
    """
    Test if the categorical dtype is preserved through the core functionality
    """

    @register(mask=["field"], demask=["field"], squeeze=["field"])
    def flagAll(saqc, field, **kwargs):
        saqc._flags[:, field] = BAD
        return saqc

    flags = initFlagsLike(data)
    flags_raw = DictOfSeries(flags)
    var1, var2 = data.columns[:2]

    pflags = SaQC(data, flags=flags_raw).flagAll(var1).flagAll(var2).flags

    for c in pflags.columns:
        assert pflags[c].dtype == flags[c].dtype


def test_new_call(data):
    qc = SaQC(data)
    qc = qc.flagRange("var1", max=5)


def test_SaQC_attributes():
    """Test if all instance attributes are in SaQC._attributes"""
    qc = SaQC()
    for name in [n for n in dir(qc) if not n.startswith("__")]:
        if hasattr(SaQC, name):  # skip class attributes
            continue
        assert name in SaQC._attributes


def test_copy(data):
    qc = SaQC(data)
    qc = qc.flagRange("var1").flagRange("var1", min=0, max=0)

    deep = qc.copy(deep=True)
    shallow = qc.copy(deep=False)
    for copy in [deep, shallow]:
        assert copy is not qc
        for name in [n for n in dir(qc) if not n.startswith("__")]:
            if hasattr(SaQC, name):  # skip class attributes
                continue
            qc_attr = getattr(qc, name)
            other_attr = getattr(copy, name)
            assert qc_attr is not other_attr

    # History is always copied
    assert deep._flags._data["var1"] is not qc._flags._data["var1"]
    assert shallow._flags._data["var1"] is not qc._flags._data["var1"]

    # underling data NOT copied
    assert shallow._data["var1"] is qc._data["var1"]

    # underling data copied
    assert deep._data["var1"] is not qc._data["var1"]


def test_sourceTargetCopy():
    """
    test implicit copies
    """
    data = initData(1)
    var1 = data.columns[0]
    target = "new"

    @register(mask=["field"], demask=["field"], squeeze=["field"], handles_target=False)
    def flagTarget(saqc, field, **kwargs):
        assert "target" not in kwargs
        return saqc

    qc = SaQC(data, flags=Flags({var1: pd.Series(127.0, index=data[var1].index)}))
    qc = qc.flagTarget(field=var1, target=target)

    assert (qc.data[var1] == qc.data[target]).all(axis=None)
    assert all(qc.flags[var1] == qc.flags[target])


def test_sourceTargetNoCopy():
    """
    test bypassing of the imlpicit copy machiners
    """
    data = initData(1)
    var1 = data.columns[0]
    target = "new"

    @register(mask=["field"], demask=["field"], squeeze=["field"], handles_target=True)
    def flagField(saqc, field, **kwargs):
        assert "target" in kwargs
        assert "target" not in saqc._data
        assert "target" not in saqc._flags
        return saqc

    SaQC(data).flagField(field=var1, target=target)


def test_sourceTargetMultivariate():
    """
    test bypassing of the imlpicit copy machinery
    """
    data = initData(3)

    @register(
        mask=["field"],
        demask=["field"],
        squeeze=["field"],
        handles_target=True,
        multivariate=True,
    )
    def flagMulti(saqc, field, **kwargs):
        assert "target" in kwargs
        assert "target" not in saqc._data
        assert "target" not in saqc._flags
        assert field == kwargs["target"]
        return saqc

    SaQC(data).flagMulti(field=data.columns, target=data.columns)


def test_sourceTargetMulti():
    data = initData(3)
    flags = initFlagsLike(data)
    fields = data.columns
    targets = [f"target{i + 1}" for i in range(len(fields))]

    @register(
        mask=["field"],
        demask=["field"],
        squeeze=["field"],
        handles_target=True,
        multivariate=True,
    )
    def flagMulti(saqc, field, target, **kwargs):
        assert len(field) == len(target)
        for src, trg in zip(field, target):
            assert src in saqc._data
            assert src in saqc._flags
            assert trg not in saqc._data
            assert trg not in saqc._flags
        return saqc

    SaQC(data, flags).flagMulti(field=fields, target=targets)


def test_unknown_attribute():
    qc = SaQC()
    with pytest.raises(AttributeError):
        qc._construct(_spam="eggs")


def test_validation(data):
    """Test if validation detects different columns in data and flags."""
    df = pd.DataFrame(
        data=np.arange(8).reshape(4, 2),
        index=pd.date_range("2020", None, 4, "1d"),
        columns=list("ab"),
    )
    qc = SaQC(df)

    @flagging()
    def flagFoo(saqc, field, **kwargs):
        saqc._data["spam"] = saqc._data[field]
        return saqc

    with pytest.raises(RuntimeError):
        qc.flagFoo("a")


def test__copy__():
    orig = SaQC()
    orig.attrs["spam"] = []  # a higher object
    shallow = copy.copy(orig)
    assert shallow is not orig
    assert shallow.attrs["spam"] is orig.attrs["spam"]


def test__deepcopy__():
    orig = SaQC()
    orig.attrs["spam"] = []  # a higher object
    shallow = copy.deepcopy(orig)
    assert shallow is not orig
    assert shallow.attrs["spam"] is not orig.attrs["spam"]


def test_immutability(data):
    field = data.columns[0]
    saqc_before = SaQC(data)
    saqc_after = saqc_before.flagDummy(field)
    for name in SaQC._attributes:
        assert getattr(saqc_before, name) is not getattr(saqc_after, name)


@pytest.mark.parametrize(
    "field,target",
    [
        (["a"], ["x", "y"]),
        (["a", "b"], ["x"]),
    ],
)
def test_fieldsTargetsExpansionFail(field, target):
    # check that the field/target handling works as expected for the
    # different function types

    @register(mask=[], demask=[], squeeze=[], multivariate=False, handles_target=False)
    def foo(saqc, field, **kwargs):
        return saqc

    @register(mask=[], demask=[], squeeze=[], multivariate=False, handles_target=True)
    def bar(saqc, field, target, **kwargs):
        return saqc

    @register(mask=[], demask=[], squeeze=[], multivariate=True, handles_target=False)
    def baz(saqc, field, **kwargs):
        return saqc

    @register(mask=[], demask=[], squeeze=[], multivariate=False, handles_target=True)
    def fooBar(saqc, field, **kwargs):
        return saqc

    data = pd.DataFrame({"a": [1, 2], "b": [2, 3], "c": [3, 4], "d": [4, 5]})
    qc = SaQC(data)
    with pytest.raises(ValueError):
        qc.foo(field, target=target)
    with pytest.raises(ValueError):
        qc.bar(field, target=target)
    with pytest.raises(ValueError):
        qc.baz(field, target=target)
    with pytest.raises(ValueError):
        qc.fooBar(field, target=target)


@pytest.mark.parametrize(
    "field,target",
    [
        (["a"], ["x"]),
        (["a", "a"], ["x", "y"]),
    ],
)
def test_fieldsTargetsExpansion(field, target):
    # check that the field/target handling works as expected for the
    # different function types

    @register(mask=[], demask=[], squeeze=[], multivariate=False, handles_target=False)
    def foo(saqc, field, **kwargs):
        return saqc

    @register(mask=[], demask=[], squeeze=[], multivariate=False, handles_target=True)
    def bar(saqc, field, target, **kwargs):
        return saqc

    @register(mask=[], demask=[], squeeze=[], multivariate=True, handles_target=False)
    def baz(saqc, field, **kwargs):
        return saqc

    @register(mask=[], demask=[], squeeze=[], multivariate=False, handles_target=True)
    def fooBar(saqc, field, **kwargs):
        return saqc

    data = pd.DataFrame({"a": [1, 2], "b": [2, 3], "c": [3, 4], "d": [4, 5]})
    qc = SaQC(data)
    qc.foo(field, target=target)
    qc.bar(field, target=target)
    qc.baz(field, target=target)
    qc.fooBar(field, target=target)


@pytest.mark.parametrize(
    "field,target",
    [
        (["a"], ["x"]),
        (["a", "a"], ["x", "y"]),
        (["a"], ["x", "y"]),
        (["a", "a"], ["x"]),
    ],
)
def test_fieldsTargetsExpansionMultivariate(field, target):
    @register(mask=[], demask=[], squeeze=[], multivariate=True, handles_target=True)
    def foo(saqc, field, target, **kwargs):
        return saqc

    @register(mask=[], demask=[], squeeze=[], multivariate=True, handles_target=True)
    def bar(saqc, field, **kwargs):
        return saqc

    data = pd.DataFrame(
        {"a": [1, 2, 3], "b": [2, 3, 4], "c": [3, 4, 5], "d": [4, 5, 6]}
    )
    qc = SaQC(data)
    qc.foo(field, target)
    qc.bar(field, target)


def test_columnConsitency(data):
    @flagging()
    def flagFoo(saqc, field, **kwargs):
        saqc._flags["spam"] = saqc._flags[field]
        return saqc

    field = data.columns[0]
    qc = SaQC(data)
    with pytest.raises(RuntimeError):
        qc.flagFoo(field)


@pytest.mark.parametrize(
    "user_flag,internal_flag",
    (
        [FILTER_ALL, FILTER_ALL],
        [FILTER_NONE, FILTER_NONE],
        [OptionalNone(), FILTER_ALL],
        ["BAD", BAD],
        ["UNFLAGGED", UNFLAGGED],
    ),
)
def test_dfilterTranslation(data, user_flag, internal_flag):
    @flagging()
    def flagFoo(saqc, field, dfilter, **kwargs):
        assert dfilter == internal_flag
        return saqc

    field = data.columns[0]
    qc = SaQC(data, scheme="simple")
    qc.flagFoo(field, dfilter=user_flag)


@pytest.mark.parametrize(
    "data, expected",
    [
        # 2c + 1c -> 3c
        (
            [
                DictOfSeries(a=pd.Series([1]), b=pd.Series([2])),
                DictOfSeries(c=pd.Series([3])),
            ],
            DictOfSeries(a=pd.Series([1]), b=pd.Series([2]), c=pd.Series([3])),
        ),
        # 1c + 1c + 1c -> 3c
        (
            [
                DictOfSeries(a=pd.Series([1])),
                DictOfSeries(b=pd.Series([2])),
                DictOfSeries(c=pd.Series([3])),
            ],
            DictOfSeries(a=pd.Series([1]), b=pd.Series([2]), c=pd.Series([3])),
        ),
    ],
)
def test_concatDios(data, expected):
    result = SaQC(data)
    assert result.data == expected


@pytest.mark.parametrize(
    "data,expected",
    [
        (
            [
                DictOfSeries(a=pd.Series([1]), b=pd.Series([2])),
                DictOfSeries(b=pd.Series([99])),
            ],
            DictOfSeries(a=pd.Series([1]), b=pd.Series([99])),
        )
    ],
)
def test_concatDios_warning(data, expected):
    with pytest.warns(UserWarning):
        result = SaQC(data)
    assert result.data == expected



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
import pandas as pd

# -*- coding: utf-8 -*-
import pytest


def test_unknownFileExtention():
    from saqc.__main__ import readData, setupIO, writeData

    reader, writer = setupIO(np.nan)
    with pytest.raises(ValueError):
        readData(reader, "foo.unknown")
    with pytest.raises(ValueError):
        writeData(reader, pd.DataFrame(), "foo.unknown")



# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later
import numpy as np
import pandas as pd
import pytest

import saqc.lib.tools as tools


@pytest.mark.parametrize("optional", [False, True])
@pytest.mark.parametrize("value", [1, 0, "foo", np.nan, np.inf])
def test_assertScalar(value, optional):
    tools.assertScalar("value", value, optional)


@pytest.mark.parametrize("optional", [False, True])
@pytest.mark.parametrize("value", [[1], [0, 1], {}, {1, 2}, pd.Series([1, 2])])
def test_assertScalar_raises(value, optional):
    with pytest.raises(ValueError):
        tools.assertScalar("value", value, optional)


def test_assertScalar_optional():
    tools.assertScalar("value", None, optional=True)
    with pytest.raises(ValueError):
        tools.assertScalar("value", None, optional=False)


class _ListLike(list):
    pass


@pytest.mark.parametrize(
    "value,expected",
    [
        (None, [None]),
        ([None], [None]),
        (1, [1]),
        (np.nan, [np.nan]),
        ([1], [1]),
        ("foo", ["foo"]),
        (["foo"], ["foo"]),
        ([1, 2], [1, 2]),
        (_ListLike("ab"), ["a", "b"]),
    ],
)
def test_toSequence(value, expected):
    result = tools.toSequence(value)
    assert isinstance(result, list)
    assert result == expected


@pytest.mark.parametrize(
    "value,expected",
    [
        # squeeze
        ([1], 1),
        ([[1, 2]], [1, 2]),
        ([[]], []),
        (_ListLike("a"), "a"),
        # no squeeze
        ([], []),
        ([1, 2], [1, 2]),
        (_ListLike("ab"), ["a", "b"]),
    ],
)
def test_squeezeSequence(value, expected):
    result = tools.squeezeSequence(value)
    assert result == expected



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-
from __future__ import annotations

import typing
from typing import TYPE_CHECKING, Callable, Tuple

import numba
import numpy as np
import pandas as pd

from saqc import BAD, UNFLAGGED
from saqc.core import DictOfSeries, Flags, flagging, register

if TYPE_CHECKING:
    from saqc import SaQC


class ChangepointsMixin:
    @flagging()
    def flagChangePoints(
        self: "SaQC",
        field: str,
        stat_func: Callable[[np.ndarray, np.ndarray], float],
        thresh_func: Callable[[np.ndarray, np.ndarray], float],
        window: str | Tuple[str, str],
        min_periods: int | Tuple[int, int],
        reduce_window: str | None = None,
        reduce_func: Callable[[np.ndarray, np.ndarray], int] = lambda x, _: x.argmax(),
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Flag values that represent a system state transition.

        Flag data points, where the parametrization of the assumed process generating this data,
        significantly changes.

        Parameters
        ----------
        stat_func :
             A function that assigns a value to every twin window. The backward-facing
             window content will be passed as the first array, the forward-facing window
             content as the second.

        thresh_func :
            A function that determines the value level, exceeding wich qualifies a
            timestamps func value as denoting a change-point.

        window :
            Size of the moving windows. This is the number of observations used for
            calculating the statistic.

            If it is a single frequency offset, it applies for the backward- and the
            forward-facing window.

            If two offsets (as a tuple) is passed the first defines the size of the
            backward facing window, the second the size of the forward facing window.

        min_periods :
            Minimum number of observations in a window required to perform the changepoint
            test. If it is a tuple of two int, the first refer to the backward-,
            the second to the forward-facing window.

        reduce_window :
            The sliding window search method is not an exact CP search method and usually
            there wont be detected a single changepoint, but a "region" of change around
            a changepoint.

            If `reduce_window` is given, for every window of size `reduce_window`, there
            will be selected the value with index `reduce_func(x, y)` and the others will
            be dropped.

            If `reduce_window` is None, the reduction window size equals the twin window
            size, the changepoints have been detected with.

        reduce_func : default argmax
            A function that must return an index value upon input of two arrays x and y.
            First input parameter will hold the result from the stat_func evaluation for
            every reduction window. Second input parameter holds the result from the
            `thresh_func` evaluation.
            The default reduction function just selects the value that maximizes the
            `stat_func`.
        """
        mask = _getChangePoints(
            data=self._data[field],
            stat_func=stat_func,
            thresh_func=thresh_func,
            window=window,
            min_periods=min_periods,
            reduce_window=reduce_window,
            reduce_func=reduce_func,
            result="mask",
        )
        self._flags[mask, field] = flag
        return self

    @register(mask=["field"], demask=[], squeeze=[])
    def assignChangePointCluster(
        self: "SaQC",
        field: str,
        stat_func: Callable[[np.ndarray, np.ndarray], float],
        thresh_func: Callable[[np.ndarray, np.ndarray], float],
        window: str | Tuple[str, str],
        min_periods: int | Tuple[int, int],
        reduce_window: str | None = None,
        reduce_func: Callable[
            [np.ndarray, np.ndarray], float
        ] = lambda x, _: x.argmax(),
        model_by_resids: bool = False,
        **kwargs,
    ) -> "SaQC":
        """
        Label data where it changes significantly.

        The labels will be stored in data. Unless `target` is given the labels will
        overwrite the data in `field`. The flags will always set to `UNFLAGGED`.

        Assigns label to the data, aiming to reflect continuous regimes of the processes
        the data is assumed to be generated by. The regime change points detection is
        based on a sliding window search.


        Parameters
        ----------
        stat_func :
            A function that assigns a value to every twin window. Left window content will
            be passed to first variable,
            right window content will be passed to the second.

        thresh_func :
            A function that determines the value level, exceeding wich qualifies a
            timestamps func value as denoting a changepoint.

        window :
            Size of the rolling windows the calculation is performed in. If it is a single
            frequency offset, it applies for the backward- and the forward-facing window.

            If two offsets (as a tuple) is passed the first defines the size of the
            backward facing window, the second the size of the forward facing window.

        min_periods :
            Minimum number of observations in a window required to perform the changepoint
            test. If it is a tuple of two int, the first refer to the backward-,
            the second to the forward-facing window.

        reduce_window :
            The sliding window search method is not an exact CP search method and usually
            there won't be detected a single changepoint, but a "region" of change around
            a changepoint. If `reduce_window` is given, for every window of size
            `reduce_window`, there will be selected the value with index `reduce_func(x,
            y)` and the others will be dropped. If `reduce_window` is None, the reduction
            window size equals the twin window size, the changepoints have been detected
            with.

        reduce_func : default argmax
            A function that must return an index value upon input of two arrays x and y.
            First input parameter will hold the result from the stat_func evaluation for
            every reduction window. Second input parameter holds the result from the
            thresh_func evaluation. The default reduction function just selects the value
            that maximizes the stat_func.

        model_by_resids :
            If True, the results of `stat_funcs` are written, otherwise the regime labels.
        """
        rtyp = "residual" if model_by_resids else "cluster"
        cluster = _getChangePoints(
            data=self._data[field],
            stat_func=stat_func,
            thresh_func=thresh_func,
            window=window,
            min_periods=min_periods,
            reduce_window=reduce_window,
            reduce_func=reduce_func,
            result=rtyp,  # type: ignore
        )
        self._data[field] = cluster
        # we set flags here against our standard policy,
        # which is not to overwrite existing flags
        self._flags[:, field] = UNFLAGGED
        return self


def _getChangePoints(
    data: pd.Series,
    stat_func: Callable[[np.ndarray, np.ndarray], float],
    thresh_func: Callable[[np.ndarray, np.ndarray], float],
    window: str | Tuple[str, str],
    min_periods: int | Tuple[int, int],
    reduce_window: str | None = None,
    reduce_func: Callable[[np.ndarray, np.ndarray], float] = lambda x, _: x.argmax(),
    result: typing.Literal["cluster", "residual", "mask"] = "mask",
) -> pd.Series:
    orig_index = data.index
    data = data.dropna()  # implicit copy

    if isinstance(window, (list, tuple)):
        bwd_window, fwd_window = window
    else:
        bwd_window = fwd_window = window

    if isinstance(window, (list, tuple)):
        bwd_min_periods, fwd_min_periods = min_periods
    else:
        bwd_min_periods = fwd_min_periods = min_periods

    if reduce_window is None:
        s = int(
            pd.Timedelta(bwd_window).total_seconds()
            + pd.Timedelta(fwd_window).total_seconds()
        )
        reduce_window = f"{s}s"

    for window in [fwd_window, bwd_window, reduce_window]:
        if isinstance(window, int):
            raise TypeError(
                "all parameter defining a size of a window "
                "must be time-offsets, not integer."
            )

    # find window bounds arrays..
    num_index = pd.Series(range(len(data)), index=data.index, dtype=int)
    # ... for the normal (backwards) case..
    rolling = num_index.rolling(bwd_window, min_periods=0)
    bwd_start = rolling.min().to_numpy(dtype=int)
    bwd_end = rolling.max().to_numpy(dtype=int) + 1
    # ... and aging for the forward case.
    rolling = num_index[::-1].rolling(fwd_window, min_periods=0, closed="left")
    fwd_start = rolling.min().fillna(len(num_index)).to_numpy(dtype=int)[::-1]
    fwd_end = (rolling.max() + 1).fillna(len(num_index)).to_numpy(dtype=int)[::-1]

    min_mask = (fwd_end - fwd_start >= fwd_min_periods) & (
        bwd_end - bwd_start >= bwd_min_periods
    )

    fwd_end = fwd_end[min_mask]
    split = bwd_end[min_mask]
    bwd_start = bwd_start[min_mask]
    masked_index = data.index[min_mask]
    check_len = len(fwd_end)
    data_arr = data.values

    # Please keep this as I sometimes need to disable jitting manually
    # to make it work with my debugger :/
    # --palmb
    try_to_jit = True
    if try_to_jit:
        jit_sf = numba.jit(stat_func, nopython=True)
        jit_tf = numba.jit(thresh_func, nopython=True)
        try:
            jit_sf(
                data_arr[bwd_start[0] : bwd_end[0]], data_arr[fwd_start[0] : fwd_end[0]]
            )
            jit_tf(
                data_arr[bwd_start[0] : bwd_end[0]], data_arr[fwd_start[0] : fwd_end[0]]
            )
            stat_func = jit_sf
            thresh_func = jit_tf
        except (numba.TypingError, numba.UnsupportedError, IndexError):
            try_to_jit = False

    args = data_arr, bwd_start, fwd_end, split, stat_func, thresh_func, check_len

    if try_to_jit:
        stat_arr, thresh_arr = _slidingWindowSearchNumba(*args)
    else:
        stat_arr, thresh_arr = _slidingWindowSearch(*args)

    result_arr = stat_arr > thresh_arr

    if result == "residuals":
        residuals = pd.Series(np.nan, index=orig_index)
        residuals[masked_index] = stat_arr
        return residuals

    det_index = masked_index[result_arr]
    detected = pd.Series(True, index=det_index)
    if reduce_window:
        length = len(detected)

        # find window bounds arrays
        num_index = pd.Series(range(length), index=detected.index, dtype=int)
        rolling = num_index.rolling(window=reduce_window, closed="both", center=True)
        start = rolling.min().to_numpy(dtype=int)
        end = (rolling.max() + 1).to_numpy(dtype=int)

        detected = _reduceCPCluster(
            stat_arr[result_arr],
            thresh_arr[result_arr],
            start,
            end,
            reduce_func,
            length,
        )
        det_index = det_index[detected]

    # The changepoint is the point "after" the change.
    # So the detected index has to be shifted by one
    # with regard to the data index.
    shifted = (
        pd.Series(True, index=det_index)
        .reindex(data.index, fill_value=False)
        .shift(fill_value=False)
    )
    det_index = shifted.index[shifted]

    mask = pd.Series(False, index=orig_index)
    mask[det_index] = True
    if result == "mask":
        return mask

    cluster = mask.cumsum()
    cluster += 1  # start cluster labels with one, not zero
    if result == "cluster":
        return cluster

    raise ValueError(
        f"'result' must be one of 'cluster', 'mask' or 'residuals' not {result}"
    )


@numba.jit(parallel=True, nopython=True)
def _slidingWindowSearchNumba(
    data_arr, bwd_start, fwd_end, split, stat_func, thresh_func, num_val
):
    stat_arr = np.zeros(num_val)
    thresh_arr = np.zeros(num_val)
    for win_i in numba.prange(0, num_val - 1):
        x = data_arr[bwd_start[win_i] : split[win_i]]
        y = data_arr[split[win_i] : fwd_end[win_i]]
        stat_arr[win_i] = stat_func(x, y)
        thresh_arr[win_i] = thresh_func(x, y)
    return stat_arr, thresh_arr


def _slidingWindowSearch(
    data_arr, bwd_start, fwd_end, split, stat_func, thresh_func, num_val
):
    stat_arr = np.zeros(num_val)
    thresh_arr = np.zeros(num_val)
    for win_i in range(0, num_val - 1):
        x = data_arr[bwd_start[win_i] : split[win_i]]
        y = data_arr[split[win_i] : fwd_end[win_i]]
        stat_arr[win_i] = stat_func(x, y)
        thresh_arr[win_i] = thresh_func(x, y)
    return stat_arr, thresh_arr


def _reduceCPCluster(stat_arr, thresh_arr, start, end, obj_func, num_val):
    out_arr = np.zeros(shape=num_val, dtype=bool)
    for win_i in numba.prange(0, num_val):
        s, e = start[win_i], end[win_i]
        x = stat_arr[s:e]
        y = thresh_arr[s:e]
        pos = s + obj_func(x, y)
        out_arr[s:e] = False
        out_arr[pos] = True

    return out_arr



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.neighbors import LocalOutlierFactor
from typing_extensions import Literal

from saqc import UNFLAGGED
from saqc.core import register
from saqc.lib.docs import DOC_TEMPLATES
from saqc.lib.tools import getApply, toSequence
from saqc.lib.ts_operators import kNN

if TYPE_CHECKING:
    from saqc import SaQC


def _kNNApply(vals, n_neighbors, func=np.sum, **kwargs):
    dist, *_ = kNN(vals, n_neighbors=n_neighbors, **kwargs)
    try:
        resids = getattr(dist, func.__name__)(axis=1)
    except AttributeError:
        resids = np.apply_along_axis(func, 1, dist)
    return resids


def _LOFApply(vals, n_neighbors, ret="scores", **kwargs):
    clf = LocalOutlierFactor(n_neighbors=n_neighbors, **kwargs)
    labels = clf.fit_predict(vals)
    resids = clf.negative_outlier_factor_
    if ret == "scores":
        return resids
    else:
        return labels == -1


def _groupedScoring(
    val_frame: pd.DataFrame,
    n: int = 20,
    freq: float | str | None = np.inf,
    min_periods: int = 2,
    score_func: Callable = _LOFApply,
    score_kwargs: Optional[dict] = None,
) -> Tuple[pd.Series, pd.Series, pd.Series]:
    score_kwargs = score_kwargs or {}
    score_index = val_frame.index
    score_ser = pd.Series(np.nan, index=score_index)

    val_frame = val_frame.loc[score_index]
    val_frame = val_frame.dropna()

    if val_frame.empty:
        return score_ser

    # partitioning
    if not freq:
        freq = val_frame.shape[0]

    if isinstance(freq, str):
        grouper = pd.Grouper(freq=freq)
    else:
        grouper = pd.Series(
            data=np.arange(0, val_frame.shape[0]), index=val_frame.index
        )
        grouper = grouper.transform(lambda x: int(np.floor(x / freq)))

    partitions = val_frame.groupby(grouper)

    for _, partition in partitions:
        if partition.empty or (partition.shape[0] < min_periods):
            continue

        sample_size = partition.shape[0]
        nn_neighbors = min(n, max(sample_size, 2) - 1)
        resids = score_func(partition.values, nn_neighbors, **score_kwargs)
        score_ser[partition.index] = resids

    return score_ser


def _univarScoring(
    data: pd.Series,
    window: Optional[str, int] = None,
    norm_func: Callable = np.nanstd,
    model_func: Callable = np.nanmean,
    center: bool = True,
    min_periods: Optional[int] = None,
) -> Tuple[pd.Series, pd.Series, pd.Series]:
    """
    Calculate (rolling) normalisation scores.

    Parameters
    ----------
    data :
        A dictionary of pandas.Series, holding all the data.
    window :
            Size of the window. Either determined via an Offset String, denoting the windows temporal extension or
            by an integer, denoting the windows number of periods.
            `NaN` measurements also count as periods.
            If `None` is passed, All data points share the same scoring window, which than equals the whole
            data.
    model_func : default std
        Function to calculate the center moment in every window.
    norm_func : default mean
        Function to calculate the scaling for every window
    center
        Weather or not to center the target value in the scoring window. If `False`, the
        target value is the last value in the window.
    min_periods
        Minimum number of valid meassurements in a scoring window, to consider the resulting score valid.
    """
    if data.empty:
        return data, data, data
    if min_periods is None:
        min_periods = 0

    if window is None:
        if data.notna().sum() >= min_periods:
            # in case of stationary analysis, broadcast statistics to series for compatibility reasons
            norm = pd.Series(norm_func(data.values), index=data.index)
            model = pd.Series(model_func(data.values), index=data.index)
        else:
            norm = pd.Series(np.nan, index=data.index)
            model = pd.Series(np.nan, index=data.index)
    else:
        # wrap passed func with rolling built in if possible and rolling.apply else
        roller = data.rolling(window=window, min_periods=min_periods, center=center)
        norm = getApply(roller, norm_func)
        model = getApply(roller, model_func)

    score = (data - model) / norm
    return score, model, norm


class ScoresMixin:
    @register(
        mask=["field"],
        demask=[],
        squeeze=["target"],
        multivariate=True,
        handles_target=True,
        docstring={"field": DOC_TEMPLATES["field"], "target": DOC_TEMPLATES["target"]},
    )
    def assignKNNScore(
        self: "SaQC",
        field: Sequence[str],
        target: str,
        n: int = 10,
        func: Callable[[pd.Series], float] = np.sum,
        freq: float | str | None = np.inf,
        min_periods: int = 2,
        algorithm: Literal["ball_tree", "kd_tree", "brute", "auto"] = "ball_tree",
        metric: str = "minkowski",
        p: int = 2,
        **kwargs,
    ) -> "SaQC":
        """
        Score datapoints by an aggregation of the dictances to their k nearest neighbors.

        The function is a wrapper around the NearestNeighbors method from pythons sklearn library (See reference [1]).

        The steps taken to calculate the scores are as follows:

        1. All the timeseries, given through ``field``, are combined to one feature space by an *inner* join on their
           date time indexes. thus, only samples, that share timestamps across all ``field`` will be included in the
           feature space.
        2. Any datapoint/sample, where one ore more of the features is invalid (=np.nan) will get excluded.
        3. For every data point, the distance to its `n` nearest neighbors is calculated by applying the
           metric `metric` at grade `p` onto the feature space. The defaults lead to the euclidian to be applied.
           If `radius` is not None, it sets the upper bound of distance for a neighbor to be considered one of the
           `n` nearest neighbors. Furthermore, the `freq` argument determines wich samples can be
           included into a datapoints nearest neighbors list, by segmenting the data into chunks of specified temporal
           extension and feeding that chunks to the kNN algorithm seperatly.
        4. For every datapoint, the calculated nearest neighbors distances get aggregated to a score, by the function
           passed to `func`. The default, ``sum`` obviously just sums up the distances.
        5. The resulting timeseries of scores gets assigned to the field target.

        Parameters
        ----------
        n : :
            The number of nearest neighbors to which the distance is comprised in every datapoints scoring calculation.

        func : default sum
            A function that assigns a score to every one dimensional array, containing the distances
            to every datapoints `n` nearest neighbors.

        freq :
            Determines the segmentation of the data into partitions, the kNN algorithm is
            applied onto individually.

            * ``np.inf``: Apply Scoring on whole data set at once
            * ``x`` > 0 : Apply scoring on successive data chunks of periods length ``x``
            * Offset String : Apply scoring on successive partitions of temporal extension matching the passed offset
              string

        min_periods :
            The minimum number of periods that have to be present in a window for the kNN scoring
            to be applied. If the number of periods present is below `min_periods`, the score for the
            datapoints in that window will be np.nan.

        algorithm :
            The search algorithm to find each datapoints k nearest neighbors.
            The keyword just gets passed on to the underlying sklearn method.
            See reference [1] for more information on the algorithm.

        metric :
            The metric the distances to any datapoints neighbors is computed with. The default of `metric`
            together with the default of `p` result in the euclidian to be applied.
            The keyword just gets passed on to the underlying sklearn method.
            See reference [1] for more information on the algorithm.

        p : :
            The grade of the metrice specified by parameter `metric`.
            The keyword just gets passed on to the underlying sklearn method.
            See reference [1] for more information on the algorithm.

        References
        ----------
        [1] https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.NearestNeighbors.html
        """
        if isinstance(target, list):
            if len(target) > 1:
                raise ValueError(
                    f"'target' must be of length 1. {target} was passed instead."
                )
            target = target[0]
        if target in self._data.columns:
            self = self.dropField(target)

        fields = toSequence(field)
        val_frame = self._data[fields].copy().to_pandas()
        score_ser = _groupedScoring(
            val_frame,
            n=n,
            freq=freq,
            min_periods=min_periods,
            score_func=_kNNApply,
            score_kwargs={
                "func": func,
                "metric": metric,
                "p": p,
                "algorithm": algorithm,
            },
        )
        score_ser.name = target

        self._flags[target] = pd.Series(np.nan, index=score_ser.index, dtype=float)
        self._data[target] = score_ser

        return self

    @register(mask=["field"], demask=[], squeeze=[])
    def assignZScore(
        self: "SaQC",
        field: str,
        window: str | None = None,
        norm_func: Callable = np.nanstd,
        model_func: Callable = np.nanmean,
        center: bool = True,
        min_periods: int | None = None,
        **kwargs,
    ) -> "SaQC":
        """
        Calculate (rolling) Zscores.

        See the Notes section for a detailed overview of the calculation

        Parameters
        ----------
        window :
            Size of the window. Either determined via an Offset String, denoting the windows temporal extension or
            by an integer, denoting the windows number of periods.
            `NaN` measurements also count as periods.
            If `None` is passed, All data points share the same scoring window, which than equals the whole
            data.
        model_func : default std
            Function to calculate the center moment in every window.
        norm_func : default mean
            Function to calculate the scaling for every window
        center :
            Weather or not to center the target value in the scoring window. If `False`, the
            target value is the last value in the window.
        min_periods :
            Minimum number of valid meassurements in a scoring window, to consider the resulting score valid.

        Notes
        -----
        Steps of calculation:

        1. Consider a window :math:`W` of successive points :math:`W = x_{1},...x_{w}`
        containing the value :math:`y_{K}` wich is to be checked.
        (The index of :math:`K` depends on the selection of the parameter `center`.)

        2. The "moment" :math:`M` for the window gets calculated via :math:`M=` `model_func(:math:`W`)

        3. The "scaling" :math:`N` for the window gets calculated via :math:`N=` `norm_func(:math:`W`)

        4. The "score" :math:`S` for the point :math:`x_{k}`gets calculated via :math:`S=(x_{k} - M) / N`
        """

        if min_periods is None:
            min_periods = 0

        score, _, _ = _univarScoring(
            self._data[field],
            window=window,
            norm_func=norm_func,
            model_func=model_func,
            center=center,
            min_periods=min_periods,
        )
        self._data[field] = score
        return self

    @register(
        mask=["field"],
        demask=[],
        squeeze=["target"],
        multivariate=True,
        handles_target=True,
        docstring={"field": DOC_TEMPLATES["field"], "target": DOC_TEMPLATES["target"]},
    )
    def assignLOF(
        self: "SaQC",
        field: Sequence[str],
        target: str,
        n: int = 20,
        freq: float | str | None = np.inf,
        min_periods: int = 2,
        algorithm: Literal["ball_tree", "kd_tree", "brute", "auto"] = "ball_tree",
        p: int = 2,
        **kwargs,
    ) -> "SaQC":
        """
        Assign Local Outlier Factor (LOF).

        Parameters
        ----------
        n :
            Number of periods to be included into the LOF calculation. Defaults to `20`, which is a value found to be
            suitable in the literature.

            * `n` determines the "locality" of an observation (its `n` nearest neighbors) and sets the upper limit of
              values of an outlier clusters (i.e. consecutive outliers). Outlier clusters of size greater than `n/2`
              may not be detected reliably.
            * The larger `n`, the lesser the algorithm's sensitivity to local outliers and small or singleton outliers
              points. Higher values greatly increase numerical costs.

        freq :
            Determines the segmentation of the data into partitions, the kNN algorithm is
            applied onto individually.
        algorithm :
            Algorithm used for calculating the `n`-nearest neighbors needed for LOF calculation.
        p :
            Degree of the metric ("Minkowski"), according to wich distance to neighbors is determined.
            Most important values are:
            * `1` - Manhatten Metric
            * `2` - Euclidian Metric
        """
        if isinstance(target, list):
            if len(target) > 1:
                raise ValueError(
                    f"'target' must be of length 1. {target} was passed instead."
                )
            target = target[0]
        if target in self._data.columns:
            self = self.dropField(target)

        fields = toSequence(field)
        val_frame = self._data[fields].copy().to_pandas()

        score_ser = _groupedScoring(
            val_frame,
            n=n,
            freq=freq,
            min_periods=min_periods,
            score_func=_LOFApply,
            score_kwargs={
                "metric": "minkowski",
                "contamination": "auto",
                "p": p,
                "algorithm": algorithm,
            },
        )
        score_ser.name = target
        self._flags[target] = pd.Series(np.nan, index=score_ser.index, dtype=float)
        self._data[target] = score_ser

        return self

    @register(mask=["field"], demask=[], squeeze=[])
    def assignUniLOF(
        self: "SaQC",
        field: str,
        n: int = 20,
        algorithm: Literal["ball_tree", "kd_tree", "brute", "auto"] = "ball_tree",
        p: int = 1,
        density: Literal["auto"] | float | Callable = "auto",
        fill_na: str = "linear",
        **kwargs,
    ) -> "SaQC":
        """
        Assign "univariate" Local Outlier Factor (LOF).

        The Function is a wrapper around a usual LOF implementation, aiming for an easy to use, parameter minimal
        outlier scoring function for singleton variables, that does not necessitate prior modelling of the variable.
        LOF is applied onto a concatenation of the `field` variable and a "temporal density", or "penalty" variable,
        that measures temporal distance between data points.

        See the Notes section for more details on the algorithm.

        Parameters
        ----------
        n :
            Number of periods to be included into the LOF calculation. Defaults to `20`, which is a value found to be
            suitable in the literature.

            * `n` determines the "locality" of an observation (its `n` nearest neighbors) and sets the upper limit of
              values of an outlier clusters (i.e. consecutive outliers). Outlier clusters of size greater than `n/2`
              may not be detected reliably.
            * The larger `n`, the lesser the algorithm's sensitivity to local outliers and small or singleton outliers
              points. Higher values greatly increase numerical costs.

        algorithm :
            Algorithm used for calculating the `n`-nearest neighbors needed for LOF calculation.

        p :
            Degree of the metric ("Minkowski"), according to wich distance to neighbors is determined.
            Most important values are:
            * `1` - Manhatten Metric
            * `2` - Euclidian Metric

        density :
            How to calculate the temporal distance/density for the variable-to-be-flagged.

            * `auto` - introduces linear density with an increment equal to the median of the absolute diff of the
              variable to be flagged
            * float - introduces linear density with an increment equal to `density`
            * Callable - calculates the density by applying the function passed onto the variable to be flagged
              (passed as Series).

        fill_na :
            Weather or not to fill NaN values in the data with a linear interpolation.

        Notes
        -----
        Algorithm steps for uniLOF flagging of variable `x`:

        1. The temporal density `dt(x)` is calculated according o the `density` parameter.
        2. LOF scores `LOF(x)` are calculated for the concatenation [`x`, `dt(x)`]
        3. `x` is flagged where `LOF(x)` exceeds the threshold determined by the parameter `thresh`.

        Examples
        --------

        """

        vals = self._data[field]
        if fill_na is not None:
            vals = vals.interpolate(fill_na)

        if density == "auto":
            density = vals.diff().abs().median()
            if density == 0:
                density = vals.diff().abs().mean()
        elif isinstance(density, Callable):
            density = density(vals)
        if isinstance(density, pd.Series):
            density = density.values

        d_var = pd.Series(np.arange(len(vals)) * density, index=vals.index)
        na_bool_ser = vals.isna() | d_var.isna()
        na_idx = na_bool_ser.index[na_bool_ser.values]
        # notna_bool = vals.notna()
        val_no = (~na_bool_ser).sum()
        if 1 < val_no < n:
            n = val_no
        elif val_no <= 1:
            return self

        d_var = d_var.drop(na_idx, axis=0).values
        vals = vals.drop(na_idx, axis=0).values
        vals_extended = np.array(
            list(vals[::-1][-n:]) + list(vals) + list(vals[::-1][:n])
        )
        d_extended = np.array(
            list(d_var[:n] + (d_var[0] - d_var[n + 1]))
            + list(d_var)
            + list(d_var[-n:] + (d_var[-1] - d_var[-n - 1]))
        )

        LOF_vars = np.array([vals_extended, d_extended]).T
        scores = _LOFApply(LOF_vars, n, p=p, algorithm=algorithm, metric="minkowski")
        scores = scores[n:-n]
        score_ser = pd.Series(scores, index=na_bool_ser.index[~na_bool_ser.values])
        score_ser = score_ser.reindex(na_bool_ser.index)
        self._data[field] = score_ser
        return self



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import TYPE_CHECKING, Tuple, Union

import numpy as np
import pandas as pd
from typing_extensions import Literal

from saqc.core import DictOfSeries, Flags, register
from saqc.lib.tools import getFreqDelta
from saqc.lib.ts_operators import (
    butterFilter,
    polyRoller,
    polyRollerIrregular,
    polyRollerNoMissing,
    polyRollerNoMissingNumba,
    polyRollerNumba,
)

if TYPE_CHECKING:
    from saqc import SaQC

FILL_METHODS = Literal[
    "linear",
    "nearest",
    "zero",
    "slinear",
    "quadratic",
    "cubic",
    "spline",
    "barycentric",
    "polynomial",
]


class CurvefitMixin:
    @register(mask=["field"], demask=[], squeeze=[])
    def fitPolynomial(
        self: "SaQC",
        field: str,
        window: int | str,
        order: int,
        min_periods: int = 0,
        **kwargs,
    ) -> "SaQC":
        """
        Fits a polynomial model to the data.

        The fit is calculated by fitting a polynomial of degree `order` to a data slice
        of size `window`, that has x at its center.

        Note that the result is stored in `field` and overwrite it unless a
        `target` is given.

        In case your data is sampled at an equidistant frequency grid:

        (1) If you know your data to have no significant number of missing values,
        or if you do not want to calculate residuals for windows containing missing values
        any way, performance can be increased by setting min_periods=window.

        Note, that the initial and final window/2 values do not get fitted.

        Each residual gets assigned the worst flag present in the interval of
        the original data.

        Parameters
        ----------
        window :
            Size of the window you want to use for fitting. If an integer is passed,
            the size refers to the number of periods for every fitting window. If an
            offset string is passed, the size refers to the total temporal extension. The
            window will be centered around the vaule-to-be-fitted. For regularly sampled
            data always a odd number of periods will be used for the fit (periods-1 if
            periods is even).

        order :
            Degree of the polynomial used for fitting

        min_periods :
            Minimum number of observations in a window required to perform the fit,
            otherwise NaNs will be assigned.
            If ``None``, `min_periods` defaults to 1 for integer windows and to the
            size of the window for offset based windows.
            Passing 0, disables the feature and will result in over-fitting for too
            sparse windows.
        """
        self._data, self._flags = _fitPolynomial(
            data=self._data,
            field=field,
            flags=self._flags,
            window=window,
            order=order,
            min_periods=min_periods,
            **kwargs,
        )
        return self

    @register(mask=["field"], demask=[], squeeze=[])
    def fitLowpassFilter(
        self: "SaQC",
        field: str,
        cutoff: float | str,
        nyq: float = 0.5,
        filter_order: int = 2,
        fill_method: FILL_METHODS = "linear",
        **kwargs,
    ) -> "SaQC":
        """
        Fits the data using the butterworth filter.

        Note
        ----
        The data is expected to be regularly sampled.

        Parameters
        ----------
        cutoff :
            The cutoff-frequency, either an offset freq string, or expressed in multiples of the sampling rate.

        nyq :
            The niquist-frequency. expressed in multiples if the sampling rate.

        fill_method :
            Fill method to be applied on the data before filtering (butterfilter cant
            handle ''np.nan''). See documentation of pandas.Series.interpolate method for
            details on the methods associated with the different keywords.

        filter_type :
            The type of filter. Default is ‘lowpass’.
        """
        self._data[field] = butterFilter(
            self._data[field],
            cutoff=cutoff,
            nyq=nyq,
            filter_order=filter_order,
            fill_method=fill_method,
            filter_type="lowpass",
        )
        return self


def _fitPolynomial(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    window: Union[int, str],
    order: int,
    min_periods: int = 0,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    # TODO: some (rather large) parts are functional similar to saqc.funcs.rolling.roll
    if data[field].empty:
        return data, flags

    to_fit = data[field].copy()
    regular = getFreqDelta(to_fit.index)
    if not regular:
        if isinstance(window, int):
            raise NotImplementedError(
                "Integer based window size is not supported for not-harmonized"
                "sample series."
            )
        # get interval centers
        centers = to_fit.rolling(
            pd.Timedelta(window) / 2, closed="both", min_periods=min_periods
        ).count()
        centers = centers.drop(centers[centers.isna()].index)
        centers = centers.astype(int)
        fitted = to_fit.rolling(
            pd.Timedelta(window), closed="both", min_periods=min_periods, center=True
        ).apply(polyRollerIrregular, args=(centers, order))
    else:
        if isinstance(window, str):
            window = pd.Timedelta(window) // regular
        if window % 2 == 0:
            window = int(window - 1)
        if min_periods is None:
            min_periods = window
        if len(to_fit) < 200000:
            numba = False
        else:
            numba = True

        val_range = np.arange(0, window)
        center_index = window // 2
        if min_periods < window:
            if min_periods > 0:
                to_fit = to_fit.rolling(
                    window, min_periods=min_periods, center=True
                ).apply(lambda x, y: x[y], raw=True, args=(center_index,))

            # we need a missing value marker that is not nan,
            # because nan values dont get passed by pandas rolling method
            miss_marker = to_fit.min()
            miss_marker = np.floor(miss_marker - 1)
            na_mask = to_fit.isna()
            to_fit[na_mask] = miss_marker
            if numba:
                fitted = to_fit.rolling(window).apply(
                    polyRollerNumba,
                    args=(miss_marker, val_range, center_index, order),
                    raw=True,
                    engine="numba",
                    engine_kwargs={"no_python": True},
                )
                # due to a tiny bug - rolling with center=True doesnt work
                # when using numba engine.
                fitted = fitted.shift(-int(center_index))
            else:
                fitted = to_fit.rolling(window, center=True).apply(
                    polyRoller,
                    args=(miss_marker, val_range, center_index, order),
                    raw=True,
                )
            fitted[na_mask] = np.nan
        else:
            # we only fit fully populated intervals:
            if numba:
                fitted = to_fit.rolling(window).apply(
                    polyRollerNoMissingNumba,
                    args=(val_range, center_index, order),
                    engine="numba",
                    engine_kwargs={"no_python": True},
                    raw=True,
                )
                # due to a tiny bug - rolling with center=True doesnt work
                # when using numba engine.
                fitted = fitted.shift(-int(center_index))
            else:
                fitted = to_fit.rolling(window, center=True).apply(
                    polyRollerNoMissing,
                    args=(val_range, center_index, order),
                    raw=True,
                )

    data[field] = fitted
    worst = flags[field].rolling(window, center=True, min_periods=min_periods).max()
    flags[field] = worst

    return data, flags



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-
from __future__ import annotations

import pickle
from typing import TYPE_CHECKING, Optional

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from typing_extensions import Literal

from saqc import FILTER_NONE, UNFLAGGED
from saqc.core import processing, register
from saqc.lib.docs import DOC_TEMPLATES
from saqc.lib.plotting import makeFig
from saqc.lib.tools import periodicMask

if TYPE_CHECKING:
    from saqc import SaQC


_MPL_DEFAULT_BACKEND = mpl.get_backend()


class ToolsMixin:
    @register(
        mask=[],
        demask=[],
        squeeze=[],
        handles_target=True,
        docstring={"target": DOC_TEMPLATES["target"]},
    )
    def copyField(
        self: "SaQC",
        field: str,
        target: str,
        overwrite: bool = False,
        **kwargs,
    ) -> "SaQC":
        """
        Copy data and flags to a new name (preserve flags history).
        """
        if field == target:
            return self

        if target in self._flags.columns.union(self._data.columns):
            if not overwrite:
                raise ValueError(f"{target}: already exist")
            self = self.dropField(field=target)

        self._data[target] = self._data[field].copy()
        self._flags.history[target] = self._flags.history[field].copy()

        return self

    @processing()
    def dropField(self: "SaQC", field: str, **kwargs) -> "SaQC":
        """
        Drops field from the data and flags.
        """
        del self._data[field]
        del self._flags[field]
        return self

    @processing()
    def renameField(self: "SaQC", field: str, new_name: str, **kwargs) -> "SaQC":
        """
        Rename field in data and flags.

        Parameters
        ----------
        new_name :
            String, field is to be replaced with.
        """
        self._data[new_name] = self._data[field]
        self._flags.history[new_name] = self._flags.history[field]
        del self._data[field]
        del self._flags[field]
        return self

    @register(mask=[], demask=[], squeeze=["field"])
    def selectTime(
        self: "SaQC",
        field: str,
        mode: Literal["periodic", "selection_field"],
        selection_field: str | None = None,
        start: str | None = None,
        end: str | None = None,
        closed: bool = True,
        **kwargs,
    ) -> "SaQC":
        """
        Realizes masking within saqc.

        Due to some inner saqc mechanics, it is not straight forwardly possible to exclude
        values or datachunks from flagging routines. This function replaces flags with UNFLAGGED
        value, wherever values are to get masked. Furthermore, the masked values get replaced by
        np.nan, so that they dont effect calculations.

        Here comes a recipe on how to apply a flagging function only on a masked chunk of the variable field:

        1. dublicate "field" in the input data (`copyField`)
        2. mask the dublicated data (this, `selectTime`)
        3. apply the tests you only want to be applied onto the masked data chunks (a saqc function)
        4. project the flags, calculated on the dublicated and masked data onto the original field data
            (`concateFlags` or `flagGeneric`)
        5. drop the dublicated data (`dropField`)

        To see an implemented example, checkout flagSeasonalRange in the saqc.functions module

        Parameters
        ----------
        mode :
            The masking mode.
            - "periodic": parameters "period_start", "end" are evaluated to generate a periodical mask
            - "mask_var": data[mask_var] is expected to be a boolean valued timeseries and is used as mask.

        selection_field :
            Only effective if mode == "mask_var"
            Fieldname of the column, holding the data that is to be used as mask. (must be boolean series)
            Neither the series` length nor its labels have to match data[field]`s index and length. An inner join of the
            indices will be calculated and values get masked where the values of the inner join are ``True``.

        start :
            Only effective if mode == "seasonal"
            String denoting starting point of every period. Formally, it has to be a truncated instance of "mm-ddTHH:MM:SS".
            Has to be of same length as `end` parameter.
            See examples section below for some examples.

        end :
            Only effective if mode == "periodic"
            String denoting starting point of every period. Formally, it has to be a truncated instance of "mm-ddTHH:MM:SS".
            Has to be of same length as `end` parameter.
            See examples section below for some examples.

        closed :
            Wheather or not to include the mask defining bounds to the mask.

        Examples
        --------
        The `period_start` and `end` parameters provide a conveniant way to generate seasonal / date-periodic masks.
        They have to be strings of the forms: "mm-ddTHH:MM:SS", "ddTHH:MM:SS" , "HH:MM:SS", "MM:SS" or "SS"
        (mm=month, dd=day, HH=hour, MM=minute, SS=second)
        Single digit specifications have to be given with leading zeros.
        `period_start` and `seas   on_end` strings have to be of same length (refer to the same periodicity)
        The highest date unit gives the period.
        For example:

        >>> start = "01T15:00:00"
        >>> end = "13T17:30:00"

        Will result in all values sampled between 15:00 at the first and  17:30 at the 13th of every month get masked

        >>> start = "01:00"
        >>> end = "04:00"

        All the values between the first and 4th minute of every hour get masked.

        >>> start = "01-01T00:00:00"
        >>> end = "01-03T00:00:00"

        Mask january and february of evcomprosed in theery year. masking is inclusive always, so in this case the mask will
        include 00:00:00 at the first of march. To exclude this one, pass:

        >>> start = "01-01T00:00:00"
        >>> end = "02-28T23:59:59"

        To mask intervals that lap over a seasons frame, like nights, or winter, exchange sequence of season start and
        season end. For example, to mask night hours between 22:00:00 in the evening and 06:00:00 in the morning, pass:

        >>> start = "22:00:00"
        >>> end = "06:00:00"
        """
        datcol_idx = self._data[field].index

        if mode == "periodic":
            mask = periodicMask(datcol_idx, start, end, ~closed)
        elif mode == "selection_field":
            idx = self._data[selection_field].index.intersection(datcol_idx)
            mask = self._data[selection_field].loc[idx]
        else:
            raise ValueError(
                "Keyword passed as masking mode is unknown ({})!".format(mode)
            )

        mask = mask.reindex(self._data[field].index, fill_value=False).astype(bool)
        self._data[field].loc[mask] = np.nan
        self._flags[mask, field] = UNFLAGGED
        return self

    @register(mask=[], demask=[], squeeze=[])
    def plot(
        self: "SaQC",
        field: str,
        path: Optional[str] = None,
        max_gap: Optional[str] = None,
        history: Optional[Literal["valid", "complete"] | list] = "valid",
        xscope: Optional[slice] = None,
        phaseplot: Optional[str] = None,
        store_kwargs: Optional[dict] = None,
        ax: mpl.axes.Axes | None = None,
        ax_kwargs: Optional[dict] = None,
        dfilter: float = FILTER_NONE,
        **kwargs,
    ) -> "SaQC":
        """
        Plot data and flags or store plot to file.

        There are two modes, 'interactive' and 'store', which are determind through the
        ``save_path`` keyword. In interactive mode (default) the plot is shown at runtime
        and the program execution stops until the plot window is closed manually. In
        store mode the generated plot is stored to disk and no manually interaction is
        needed.

        Parameters
        ----------
        path :
            If ``None`` is passed, interactive mode is entered; plots are shown immediatly
            and a user need to close them manually before execution continues.
            If a filepath is passed instead, store-mode is entered and
            the plot is stored unter the passed location.

        max_gap :
            If None, all the points in the data will be connected, resulting in long linear
            lines, where continous chunks of data is missing. Nans in the data get dropped
            before plotting. If an offset string is passed, only points that have a distance
            below `max_gap` get connected via the plotting line.

        history :
            Discriminate the plotted flags with respect to the tests they originate from.

            * "valid" - Only plot those flags, that do not get altered or "unflagged" by subsequent tests. Only list tests
              in the legend, that actually contributed flags to the overall resault.
            * "complete" - plot all the flags set and list all the tests ran on a variable. Suitable for debugging/tracking.
            * None - just plot the resulting flags for one variable, without any historical meta information.
            * list of strings - plot only flags set by those tests listed.

        xscope :
            Parameter, that determines a chunk of the data to be plotted
            processed. `xscope` can be anything, that is a valid argument to the ``pandas.Series.__getitem__`` method.

        phaseplot :
            If a string is passed, plot ``field`` in the phase space it forms together with the Variable ``phaseplot``.

        store_kwargs :
            Keywords to be passed on to the ``matplotlib.pyplot.savefig`` method, handling
            the figure storing. To store an pickle object of the figure, use the option
            ``{'pickle': True}``, but note that all other store_kwargs are ignored then.
            Reopen with: ``pickle.load(open(savepath,'w')).show()``

        ax_kwargs :
            Axis keywords. Change the axis labeling defaults. Most important keywords:
            'x_label', 'y_label', 'title', 'fontsize', 'cycleskip'.
        """
        data, flags = self._data.copy(), self._flags.copy()

        level = kwargs.get("flag", UNFLAGGED)

        if dfilter < np.inf:
            data[field].loc[flags[field] >= dfilter] = np.nan

        if store_kwargs is None:
            store_kwargs = {}

        if ax_kwargs is None:
            ax_kwargs = {}

        if not path:
            mpl.use(_MPL_DEFAULT_BACKEND)
        else:
            mpl.use("Agg")

        fig = makeFig(
            data=data,
            field=field,
            flags=flags,
            level=level,
            max_gap=max_gap,
            history=history,
            xscope=xscope,
            phaseplot=phaseplot,
            ax=ax,
            ax_kwargs=ax_kwargs,
        )

        if ax is None and not path:
            plt.show()

        if path:
            if store_kwargs.pop("pickle", False):
                with open(path, "wb") as f:
                    pickle.dump(fig, f)
            else:
                fig.savefig(path, **store_kwargs)

        return self



#! /usr/bin/env python
# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
# SPDX-License-Identifier: GPL-3.0-or-later
# -*- coding: utf-8 -*-

# isort: skip_file
from saqc.core.frame import DictOfSeries
from saqc.core.history import History
from saqc.core.flags import Flags, initFlagsLike
from saqc.core.register import flagging, processing, register
from saqc.core.core import SaQC
